from dataclasses import dataclass
from typing import List, Optional

from dolma.cli import BaseCli, field, print_config
from dolma.cli.shared import WorkDirConfig, make_workdirs
from dolma.core.errors import DolmaConfigError
from dolma.core.loggers import get_logger
from dolma.core.paths import glob_path
from dolma.tokenizer import tokenize_in_parallel


@dataclass
class TokenizerConfig:
    name_or_path: Optional[str] = field(
        default=None,
        help="Name or path of the tokenizer to use. Must be a HuggingFace-compatible tokenizer. Required.",
    )
    bos_token_id: Optional[int] = field(
        default=None, help="The token ID corresponding to the 'beginning-of-sentence' token."
    )
    eos_token_id: Optional[int] = field(
        default=None,
        help="The token ID corresponding to the 'end-of-sentence' token.",
    )
    pad_token_id: Optional[int] = field(
        default=None,
        help="The token ID corresponding to the 'padding' token.",
    )
    segment_before_tokenization: bool = field(
        default=False,
        help=(
            "Whether to segment documents by paragraph before tokenization. "
            "This is useful for tokenizers like Llama that are very slow on long documents. "
            "Might not be needed once this bugfix is merged https://github.com/huggingface/tokenizers/pull/1413"
        ),
    )

    def __post__init__(self):
        logger = get_logger(__file__)

        if self.eos_token_id is None:
            logger.warning("NO EOS TOKEN PROVIDED. Are you sure this is what you want?")

        if self.bos_token_id is None:
            logger.warning("NO BOS TOKEN PROVIDED. Are you sure this is what you want?")

        if self.pad_token_id is None:
            logger = get_logger(__file__)
            logger.warning("No pad token ID provided; using EOS token ID.")
            self.pad_token_id = self.eos_token_id

        if self.segment_before_tokenization:
            logger.warning(
                "EXPERIMENTAL FEATURE: segmenting before tokenization is enabled. "
                "This option has only been tested with Llama and GPT-NeoX tokenizers. "
                "USE AT YOUR OWN RISK."
            )

    @classmethod
    def deprecated_init(cls, tokenizer_name_or_path: str) -> "TokenizerConfig":
        logger = get_logger(__file__)
        logger.warning(
            "The `tokenizer_name_or_path` argument is deprecated and will be removed in a future release. "
            "Please use --tokenizer.name_or_path, and provide --tokenizer.eos_token_id as well "
            "(and, optionally, --tokenizer.pad_token_id)."
        )
        from tokenizers import Tokenizer

        # before options to pass eos_token_id and pad_token_id were added, eos was set to
        # the last token in the vocab. We need to do the same here to maintain compatibility
        legacy_tokenizer = Tokenizer.from_pretrained(tokenizer_name_or_path)
        old_eos_token_id = len(legacy_tokenizer.get_vocab()) - 1

        return cls(
            name_or_path=tokenizer_name_or_path,
            eos_token_id=old_eos_token_id,
            segment_before_tokenization=False,
        )


@dataclass
class TokenizationConfig:
    documents: List[str] = field(
        default=[],
        help=(
            "One or more document paths to process; Can be either local or S3 paths. "
            "Globs are supported. Required"
        ),
    )
    destination: Optional[str] = field(
        default=None,
        help=(
            "Destination paths to save the outputs; should match the number of document paths. "
            "If not provided, destination will be derived from the document path. Required."
        ),
    )
    tokenizer_name_or_path: Optional[str] = field(
        default=None,
        help="Deprecated. Use --tokenizer.name_or_path instead.",
    )
    tokenizer: Optional[TokenizerConfig] = field(
        default=None,
        help="Configuration for the tokenizer.",
    )
    processes: int = field(
        default=1,
        help="Number of parallel processes to use.",
    )
    files_per_process: Optional[int] = field(
        default=None,
        help="Number of files to process per process.",
    )
    batch_size: int = field(
        default=10_000,
        help="Number of sequences to tokenize before writing to disk.",
    )
    ring_size: int = field(default=8, help="Number of files to open in parallel for tokenization.")
    max_size: int = field(
        default=1024 * 1024 * 1024,
        help="Maximum size of a file in bytes.",
    )
    dtype: str = field(
        default="uint16",
        help="Data type for the memmap file; must be a valid numpy dtype.",
    )
    debug: bool = field(
        default=False,
        help="Whether to run in debug mode.",
    )
    seed: int = field(
        default=3920,
        help="Seed for random number generation.",
    )
    work_dir: WorkDirConfig = field(default=WorkDirConfig(), help="Configuration for temporary work directories.")
    dryrun: bool = field(
        default=False,
        help="If true, only print the configuration and exit without running the taggers.",
    )


class TokenizerCli(BaseCli):
    CONFIG = TokenizationConfig
    DESCRIPTION = "Tokenize documents using the provided tokenizer."

    @classmethod
    def run(cls, parsed_config: TokenizationConfig):
        logger = get_logger("tagger")

        with make_workdirs(parsed_config.work_dir) as work_dirs:
            documents = [str(p) for p in parsed_config.documents]

            # perform some path validation to make sure we don't call the mixer with invalid config
            total_matching_documents = 0
            for document in documents:
                current_matching_documents = sum(1 for _ in glob_path(document))
                if current_matching_documents == 0:
                    # only raise a warning if no documents are found for a single path
                    logger.warning("No documents found for path %s", document)
                total_matching_documents += current_matching_documents

            if total_matching_documents == 0:
                # but raise an error if no documents are found for all paths
                raise DolmaConfigError(f"No documents found for paths {documents}.")

            print_config(parsed_config)
            if parsed_config.dryrun:
                logger.info("Exiting due to dryrun.")
                return

            if parsed_config.destination is None:
                raise DolmaConfigError("Destination must be provided.")

            # must handle new and deprecated way to get tokenizer config
            if parsed_config.tokenizer is None:
                if parsed_config.tokenizer_name_or_path is None:
                    raise DolmaConfigError("Tokenizer configuration is missing.")
                else:
                    parsed_config.tokenizer = TokenizerConfig.deprecated_init(parsed_config.tokenizer_name_or_path)

            if parsed_config.tokenizer.name_or_path is None:
                raise DolmaConfigError("Tokenizer name or path must be provided.")

            tokenize_in_parallel(
                sources=documents,
                destination=parsed_config.destination,
                num_writers=parsed_config.processes,
                num_readers=parsed_config.files_per_process,
                local_shuffle=parsed_config.batch_size,
                ring_size=parsed_config.ring_size,
                tokenizer_name_or_path=parsed_config.tokenizer.name_or_path,
                eos_token_id=parsed_config.tokenizer.eos_token_id,
                pad_token_id=parsed_config.tokenizer.pad_token_id,
                segment_before_tokenization=parsed_config.tokenizer.segment_before_tokenization,
                seed=parsed_config.seed,
                metadata_dir=work_dirs.output,
                max_size=parsed_config.max_size,
                debug=parsed_config.debug,
            )



from typing import List, NamedTuple

from ..core.data_types import InputSpec

__all__ = ["InputSpec", "TokenizerOutput", "Metadata"]


class TokenizerOutput(NamedTuple):
    id: str
    src: str
    loc: int
    tokens: List[int]
    start: int
    end: int

    @classmethod
    def from_tokens(cls, id: str, src: str, loc: int, tokens: List[int]) -> "TokenizerOutput":
        return cls(id=id, src=src, loc=loc, tokens=tokens, start=0, end=len(tokens))

    @classmethod
    def from_output_spec(cls, output_spec: "TokenizerOutput", start: int = -1, end: int = -1) -> "TokenizerOutput":
        start = start if start >= 0 else output_spec.start
        end = end if end >= 0 else output_spec.end
        return cls(
            id=output_spec.id,
            src=output_spec.src,
            loc=output_spec.loc,
            tokens=output_spec.tokens,
            start=start,
            end=end,
        )


class Metadata(NamedTuple):
    id: str
    src: str
    loc: int
    start: int
    end: int

    def to_csv(self) -> str:
        return f"{self.id},{self.src},{self.loc},{self.start},{self.end}"


class MemmapMetadata(NamedTuple):
    start: int
    end: int
    id: str
    src: str
    loc: int



from dataclasses import dataclass
from typing import List, Optional

from dolma.cli import BaseCli, field, print_config
from dolma.cli.shared import WorkDirConfig, make_workdirs
from dolma.core.analyzer import create_and_run_analyzer
from dolma.core.errors import DolmaConfigError
from dolma.core.loggers import get_logger
from dolma.core.paths import glob_path


@dataclass
class AnalyzerConfig:
    attributes: List[str] = field(
        default=[],
        help="One or more attributes paths to process; Can be either local or S3 paths. Globs are supported.",
    )
    report: Optional[str] = field(
        default=None,
        help=(
            "Path where to save the report. Can be either local or S3 path. "
            "If not provided, the report will be printed to stdout."
        ),
    )
    bins: int = field(
        default=1_000,
        help="Number of bins to use for the histograms.",
    )
    processes: int = field(
        default=1,
        help="Number of parallel processes to use.",
    )
    seed: int = field(
        default=0,
        help="Seed to use for reproducibility.",
    )
    debug: bool = field(
        default=False,
        help="Whether to run in debug mode.",
    )
    work_dir: WorkDirConfig = field(default=WorkDirConfig(), help="Configuration for temporary work directories.")
    regex: Optional[str] = field(
        default=None,
        help="Regex to use for filtering the attributes by name.",
    )


class AnalyzerCli(BaseCli):
    CONFIG = AnalyzerConfig
    DESCRIPTION = "Analyze the distribution of attributes values in a dataset."

    @classmethod
    def run(cls, parsed_config: AnalyzerConfig):
        logger = get_logger("analyzer")

        # perform some path validation to make sure we don't call the mixer with invalid config
        total_matching_documents = 0
        for document in parsed_config.attributes:
            current_matching_documents = sum(1 for _ in glob_path(document))
            if current_matching_documents == 0:
                # only raise a warning if no documents are found for a single path
                logger.warning("No documents found for path %s", document)
            total_matching_documents += current_matching_documents

        if total_matching_documents == 0:
            # but raise an error if no documents are found for all paths
            raise DolmaConfigError(f"No documents found for paths {parsed_config.attributes}.")

        print_config(parsed_config)

        with make_workdirs(parsed_config.work_dir) as work_dirs:
            create_and_run_analyzer(
                attributes=parsed_config.attributes,
                report=parsed_config.report,
                summaries_path=work_dirs.output,
                metadata_path=work_dirs.input,
                debug=parsed_config.debug,
                seed=parsed_config.seed,
                num_bins=parsed_config.bins,
                num_processes=parsed_config.processes,
                name_regex=parsed_config.regex,
            )



import sys
import unittest
from pathlib import Path

from dolma.core.registry import TaggerRegistry
from dolma.core.utils import import_modules


class TestExtra(unittest.TestCase):
    def setUp(self) -> None:
        self.current_path = Path(__file__).parent.absolute()

    def test_import_from_module(self):
        sys.path.append(f"{self.current_path}/extras")
        import_modules(["extras_from_module"])
        self.assertTrue(TaggerRegistry.has("extra_v1"))

    def test_import_from_path(self):
        import_modules([f"{self.current_path}/extras/extras_from_path/extra_taggers.py"])
        self.assertTrue(TaggerRegistry.has("extra_v2"))

    def test_import_from_module_path(self):
        import_modules([f"{self.current_path}/extras/extras_from_module_path"])
        self.assertTrue(TaggerRegistry.has("extra_v3"))



import regex

from ..core.data_types import DocResult, Document, Span
from ..core.registry import TaggerRegistry
from ..core.taggers import BaseTagger
from ..core.utils import split_paragraphs


@TaggerRegistry.add("not_alphanum_paragraph_v1")
class NotAlphanumParagraphV1(BaseTagger):
    def __init__(self) -> None:
        self.re_has_alphanum = regex.compile(r"[a-zA-Z0-9]", regex.UNICODE)
        self.re_all_punctuation = regex.compile(
            r"^("
            r"[[:punct:]]|"
            r"\s|"
            r"["
            "\U0001F300-\U0001F64F"
            "\U0001F680-\U0001F6FF"
            "\u2600-\u26FF\u2700-\u27BF"
            r"]+"
            r")+$",
            regex.UNICODE,
        )

    def predict(self, doc: Document) -> DocResult:
        spans = []

        for para in split_paragraphs(text=doc.text):
            if self.re_has_alphanum.search(para.text):
                continue

            if self.re_all_punctuation.search(para.text):
                spans.append(Span(start=para.start, end=para.end, type="all_punct", score=1))

        if not spans:
            spans.append(Span(start=0, end=len(doc.text), type="all_punct", score=0))

        return DocResult(doc=doc, spans=spans)



"""

Data types assumed by Filters.

@kylel, @soldni

"""

from typing import Any, Dict, List, Optional, Tuple

from msgspec import Struct
from typing_extensions import TypeAlias

TaggerOutputValueType: TypeAlias = Tuple[int, int, float]
TaggerOutputType: TypeAlias = List[TaggerOutputValueType]
TaggerOutputDictType: TypeAlias = Dict[str, TaggerOutputType]


class InputSpec(Struct):
    id: str
    text: str
    source: str = ""
    version: Optional[str] = None


class InputSpecWithMetadata(InputSpec):
    metadata: Optional[Dict[str, Any]] = None


class OutputSpec(Struct):
    id: str
    attributes: Dict[str, List[Tuple[int, int, float]]]
    source: Optional[str] = None


class Document:
    __slots__ = "source", "version", "id", "text"

    def __init__(self, source: str, id: str, text: str, version: Optional[str] = None) -> None:
        self.source = source
        self.version = version
        self.id = id
        self.text = text

    @classmethod
    def from_spec(cls, spec: InputSpec) -> "Document":
        return Document(source=spec.source, version=spec.version, id=spec.id, text=spec.text)

    def to_spec(self) -> InputSpec:
        return InputSpec(source=self.source, version=self.version, id=self.id, text=self.text)

    @classmethod
    def from_json(cls, d: Dict[str, Any]) -> "Document":
        return Document(source=d["source"], version=d["version"], id=d["id"], text=d["text"])

    def to_json(self) -> Dict[str, Any]:
        return {"source": self.source, "version": self.version, "id": self.id, "text": self.text}

    def __str__(self) -> str:
        attributes_string = ",".join([f"{k}:{repr(v)}" for k, v in self.to_json().items()])
        return f"{self.__class__.__name__}({attributes_string})"


class DocumentWithMetadata(Document):
    __slots__ = ("metadata",)

    def __init__(self, *args, metadata: Optional[Dict[str, Any]] = None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.metadata = metadata or {}

    @classmethod
    def from_spec(cls, spec: InputSpecWithMetadata) -> "DocumentWithMetadata":
        return DocumentWithMetadata(
            source=spec.source,
            version=spec.version,
            id=spec.id,
            text=spec.text,
            metadata=spec.metadata,
        )

    def to_spec(self) -> InputSpecWithMetadata:
        return InputSpecWithMetadata(
            source=self.source,
            version=self.version,
            id=self.id,
            text=self.text,
            metadata=self.metadata,
        )

    @classmethod
    def from_json(cls, d: Dict) -> "DocumentWithMetadata":
        return DocumentWithMetadata(
            source=d["source"],
            version=d["version"],
            id=d["id"],
            text=d["text"],
            metadata=d["metadata"],
        )

    def to_json(self) -> Dict:
        return {
            "source": self.source,
            "version": self.version,
            "id": self.id,
            "text": self.text,
            "metadata": self.metadata,
        }

    def __str__(self) -> str:
        repr_ = super().__str__()
        return repr_.rstrip(")") + f",metadata={'...' if self.metadata else 'none'})"


class Span:
    __slots__ = "start", "end", "type", "score", "experiment", "tagger"

    def __init__(
        self,
        start: int,
        end: int,
        type: str,
        score: float = 1.0,
        experiment: Optional[str] = None,
        tagger: Optional[str] = None,
    ):
        self.start = start
        self.end = end
        self.type = type
        self.score = float(score)
        self.experiment = experiment
        self.tagger = tagger

    def mention(self, text: str, window: int = 0) -> str:
        return text[max(0, self.start - window) : min(len(text), self.end + window)]

    def select(self, doc: Document) -> str:
        return doc.text[self.start : self.end]

    @classmethod
    def from_spec(cls, attribute_name: str, attribute_value: TaggerOutputValueType) -> "Span":
        if "__" in attribute_name:
            # bff tagger has different name
            exp_name, tgr_name, attr_type = attribute_name.split("__", 2)
        else:
            exp_name = tgr_name = attr_type = attribute_name

        start, end, score = attribute_value
        return Span(
            start=int(start),
            end=int(end),
            type=attr_type,
            score=float(score),
            experiment=exp_name,
            tagger=tgr_name,
        )

    def to_spec(self) -> Tuple[str, TaggerOutputValueType]:
        assert self.experiment is not None, "Experiment name must be set to convert to spec"
        assert self.tagger is not None, "Tagger name must be set to convert to spec"
        return (
            f"{self.experiment}__{self.tagger}__{self.type}",
            (self.start, self.end, self.score),
        )

    def __len__(self) -> int:
        return self.end - self.start

    @classmethod
    def from_json(cls, di: Dict) -> "Span":
        return Span(start=di["start"], end=di["end"], type=di["type"], score=di["score"])

    def to_json(self, text: Optional[str] = None, window: int = 0) -> dict:
        span_repr = {"start": self.start, "end": self.end, "type": self.type, "score": self.score}
        if text is not None:
            span_repr["mention"] = self.mention(text=text, window=window)
        return span_repr

    def __str__(self) -> str:
        cls_name = self.__class__.__name__
        return f"{cls_name}(start={self.start},end={self.end},type={repr(self.type)},score={self.score:.5f})"

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return (
            self.start == other.start
            and self.end == other.end
            and self.type == other.type
            and self.score == other.score
        )


class DocResult:
    __slots__ = "doc", "spans"

    def __init__(self, doc: Document, spans: List[Span]) -> None:
        self.doc = doc
        self.spans = spans

    @classmethod
    def from_spec(cls, doc: InputSpec, *attrs_groups: OutputSpec) -> "DocResult":
        spans: List[Span] = []
        for attrs in attrs_groups:
            assert doc.id == attrs.id, f"doc.id={doc.id} != attrs.id={attrs.id}"
            spans.extend(
                [
                    Span.from_spec(attribute_name=attr_name, attribute_value=attr_value)
                    for attr_name, attr_values in attrs.attributes.items()
                    for attr_value in attr_values
                ]
            )
        return DocResult(doc=Document.from_spec(doc), spans=spans)

    def to_spec(self) -> Tuple[InputSpec, OutputSpec]:
        doc_spec = self.doc.to_spec()
        attributes: Dict[str, List[TaggerOutputValueType]] = {}

        for span in self.spans:
            attr_name, attr_value = span.to_spec()
            attributes.setdefault(attr_name, []).append(attr_value)

        return doc_spec, OutputSpec(source=self.doc.source, id=self.doc.id, attributes=attributes)

    @classmethod
    def from_json(cls, d: Dict[str, Any]) -> "DocResult":
        return DocResult(
            doc=Document.from_json(d["doc"]),
            spans=[Span.from_json(span) for span in d["spans"]],
        )

    def to_json(self, with_doc: bool = False, window: int = 0) -> Dict[str, Any]:
        d: Dict[str, Any] = {"spans": [span.to_json(text=self.doc.text, window=window) for span in self.spans]}
        if with_doc:
            d["doc"] = self.doc.to_json()
        return d

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(doc={self.doc},spans=[{','.join(str(s) for s in self.spans)}])"


class TextSlice:
    """A slice of text from a document."""

    __slots__ = "doc", "start", "end"

    def __init__(self, doc: str, start: int, end: int):
        self.doc = doc
        self.start = start
        self.end = end

    @property
    def text(self) -> str:
        return self.doc[self.start : self.end]

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(text={repr(self.text)},start={self.start},end={self.end})"



import argparse
import bisect
import fnmatch
import re
from collections import defaultdict
from functools import partial
from pathlib import Path
from statistics import stdev
from typing import List, Optional, Sequence, Tuple

from necessary import necessary

with necessary(
    modules=["plotly", "wandb", "kaleido", "yaml"],
    message="Please run `pip install plotly wandb kaleido pyyaml`",
):
    import plotly.graph_objs as go
    import plotly.io as pio
    import yaml

    import wandb

    pio.kaleido.scope.mathjax = None


def parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument("-t", "--wandb-team", type=str, required=True, help="Name of the W&B team")
    ap.add_argument("-p", "--wandb-project", type=str, required=True, help="Name of the W&B project")
    ap.add_argument("-n", "--wandb-names", type=str, nargs="+", required=True, help="Run name or regex")
    ap.add_argument("-x", "--x-axis", type=str, default="throughput/total_tokens", help="X axis")
    ap.add_argument("-y", "--y-axis", nargs="+", type=str, default=["train/Perplexity"], help="Y axis")
    ap.add_argument("-s", "--samples", default=0, type=int, help="Samples to get; 0 for no sampling")
    ap.add_argument("-d", "--destination", type=Path, required=True, help="Directory to save plots to")
    ap.add_argument("-e", "--plot-extension", type=str, default="pdf", help="Plot extension to use")
    ap.add_argument("--max-x-axis", type=float, default=None, help="Maximum x value")
    ap.add_argument("--max-y-axis", type=float, default=None, help="Maximum y value")
    ap.add_argument("--remove-outliers", action="store_true", help="Remove outliers")
    ap.add_argument("--legend-title", type=str, default=None, help="Legend title")
    ap.add_argument("--y-log-scale", action="store_true", help="Use log scale for y axis")
    ap.add_argument("-v", "--vocabulary", type=Path, default=None, help="Path json file with pretty names")
    ap.add_argument("-N", "--experiment-nickname", type=str, default=None, help="Experiment nickname")
    ap.add_argument("--plotly-theme", type=str, default="none", help="Plotly theme to use")
    ap.add_argument("--plotly-font-size", type=int, default=10, help="Plotly font size")
    ap.add_argument("--plotly-show-title", action="store_true", help="Show plot title")
    ap.add_argument("--plotly-figure-width", type=int, default=800, help="Plotly figure width")
    ap.add_argument("--plotly-figure-height", type=int, default=500, help="Plotly figure height")
    return ap.parse_args()


def match_run_name(name: str, run_names: List[str]) -> Optional[str]:
    for run_name in run_names:
        if fnmatch.filter([name], run_name) or re.search(run_name, name):
            return run_name
    return None


def remove_outliers(
    x: Sequence[float], y: Sequence[float], z: float = 10.0
) -> Tuple[Sequence[float], Sequence[float]]:
    if min(y) >= 0 and max(y) <= 1:
        # do not crop values that are in [0, 1]
        return x, y

    std = stdev(y)
    x, y = zip(*[(x, y) for x, y in zip(x, y) if abs(y - std) < z * std])
    return x, y


def main():
    opts = parse_args()

    # make sure we're logged in
    wandb.login()

    # get all the runs matching name filters
    api = wandb.Api()
    wb_path = f"{opts.wandb_team}/{opts.wandb_project}"
    # wb_filters = {"$or": [{"config.run_name": {"$regex": n}} for n in opts.wandb_names]}
    wb_filters = {"$or": [{"display_name": {"$regex": n}} for n in opts.wandb_names]}
    wb_runs = api.runs(path=wb_path, filters=wb_filters)

    vocabulary = {}
    if opts.vocabulary is not None and opts.vocabulary.exists():
        with opts.vocabulary.open("r") as f:
            vocabulary = yaml.safe_load(f.read())

    # prepare the vocabulary by (a) overriding with experiment specific
    # names that are not in the vocabulary file and (b) experiment config
    # (nested vocabulary)
    vocabulary = {
        **{k: v for k, v in vocabulary.items() if isinstance(v, str)},
        **{k: v for k, v in vocabulary.get(opts.experiment_nickname, {}).items() if isinstance(v, str)},
    }

    metrics = defaultdict(lambda: {n: {"x": [], "y": []} for n in opts.wandb_names})
    run_name_matcher = partial(match_run_name, run_names=opts.wandb_names)

    print(f"Found {len(wb_runs)} matching runs in {wb_path}")

    for wb_run in wb_runs:
        plot_group_name = run_name_matcher(wb_run.name)
        if plot_group_name is None:
            raise ValueError(f"Could not find a name match for {wb_run.name}")

        print(f"Processing run {wb_run.name} into group {plot_group_name}")

        if opts.samples > 0:
            x_axis_history = wb_run.history(samples=opts.samples, keys=[opts.x_axis], pandas=False)
        else:
            x_axis_history = list(wb_run.scan_history(keys=[opts.x_axis]))

        if len(x_axis_history) == 0:
            # this run has crashed and has no history
            print(f"WARNING: skipping {wb_run.name} because it has no history. Crashed early?")
            continue

        steps, x_axis = zip(*sorted([(wb_step["_step"], wb_step[opts.x_axis]) for wb_step in x_axis_history]))

        if opts.samples > 0:
            history = wb_run.history(samples=opts.samples, keys=opts.y_axis, pandas=False)
        else:
            history = wb_run.scan_history(keys=opts.y_axis)

        for y_axis in opts.y_axis:
            yaxis_pretty_name = vocabulary.get(y_axis, y_axis)
            for wb_step in history:
                loc = min(bisect.bisect_left(steps, wb_step["_step"]), len(x_axis) - 1)
                metrics[yaxis_pretty_name][plot_group_name]["x"].append(x_axis[loc])
                metrics[yaxis_pretty_name][plot_group_name]["y"].append(wb_step[y_axis])

    xaxis_pretty_name = vocabulary.get(opts.x_axis, opts.x_axis)

    for y_axis, plot_groups in metrics.items():
        fig = go.Figure()

        # these we figure out as we go
        use_y_log = opts.y_log_scale
        top_right_legend = False

        for run_name, run_data in plot_groups.items():
            if len(run_data["y"]) == 0:
                print(f"WARNING: skipping {run_name} because it has no data for {y_axis}")
                continue

            # start by sorting the data by x axis
            x, y = zip(*sorted(zip(run_data["x"], run_data["y"])))

            if opts.remove_outliers:
                x, y = remove_outliers(x=x, y=y)

            if opts.max_x_axis and min(x) < opts.max_x_axis:
                x, y = zip(*[(x, y) for x, y in zip(x, y) if x <= opts.max_x_axis])

            if opts.max_y_axis and min(y) < opts.max_y_axis:
                x, y = zip(*[(x, y) for x, y in zip(x, y) if y <= opts.max_y_axis])

            if y[-1] < y[0]:
                # decreasing y values, so we want the legend on the top right
                top_right_legend = True

            # only use log scale if we have values that are not in [0, 1]
            if max(y) > 1 or min(y) < 0:
                min_y = min([y for y in y if y > 0] or [1e-3])  # avoid diving by zero
                use_y_log = use_y_log or (max(y) / min_y > 100)

            figure_run_name = vocabulary.get(run_name, run_name)
            fig.add_trace(go.Scatter(name=figure_run_name, x=x, y=y, mode="lines"))

        legend_config = {
            "yanchor": "top" if top_right_legend else "bottom",
            "y": 0.99 if top_right_legend else 0.01,
            "xanchor": "right",
            "x": 0.99,
            "font": {"size": opts.plotly_font_size},
            "borderwidth": 1,
            "bordercolor": "Gray",
        }
        fig.update_layout(legend=legend_config)

        title_text = vocabulary.get(opts.experiment_nickname, opts.experiment_nickname)
        fig.update_layout(
            template=opts.plotly_theme,
            xaxis_title=xaxis_pretty_name,
            yaxis_title=y_axis,
            legend_title=opts.legend_title,
            title_text=title_text if opts.plotly_show_title else None,
            font=dict(size=opts.plotly_font_size),
            width=opts.plotly_figure_width,
            height=opts.plotly_figure_height,
            margin=dict(
                l=4 * opts.plotly_font_size,
                r=opts.plotly_font_size,
                b=4 * opts.plotly_font_size,
                t=opts.plotly_font_size,
            ),
        )
        fig.update_yaxes(type="log" if use_y_log else "linear")
        fig.update_xaxes(range=([0, opts.max_x_axis] if opts.max_x_axis is not None else None))

        file_name = re.sub(r"\W+", "_", y_axis).lower().strip("_")
        file_path = opts.destination / f"{file_name}.{opts.plot_extension}"
        file_path.parent.mkdir(parents=True, exist_ok=True)
        fig.write_image(str(file_path))
        print(f"Saved {y_axis} plot to {file_path}")


NORMALIZE_PLOTLY_NAME = re.compile(r"([a-z])([A-Z])")

if __name__ == "__main__":
    main()



from .data_types import TokenizerOutput
from .executor import tokenize_in_parallel
from .tokenizer import Tokenizer, tokenize_file

__all__ = [
    "Tokenizer",
    "tokenize_file",
    "tokenize_in_parallel",
    "TokenizerOutput",
]



import json
import os
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import TestCase

import smart_open

from dolma.core.runtime import (
    _make_paths_from_prefix,
    _make_paths_from_substitution,
    create_and_run_tagger,
)

LOCAL_DATA = Path(__file__).parent.parent / "data"


class TestRuntimeUtilities(TestCase):
    def test_make_paths_from_substitution(self):
        paths = [
            "s3://bucket/common-crawl/documents/cc_*/*.json.gz",
            "/local/path/to/documents/train/*",
        ]
        new_paths = _make_paths_from_substitution(
            paths=paths,
            find="documents",
            replace="attributes",
        )
        self.assertEqual(new_paths, ["s3://bucket/common-crawl/attributes", "/local/path/to/attributes/train"])

    def test_make_paths_from_prefix(self):
        paths = [
            "s3://bucket/common-crawl/documents/cc_head/*.json.gz",
            "s3://bucket/common-crawl/documents/cc_middle/*.json.gz",
            "s3://bucket/common-crawl/documents/cc_tail/*.json.gz",
        ]
        new_paths = _make_paths_from_prefix(
            paths=paths,
            prefix="s3://bucket/common-crawl/attributes/",
        )
        self.assertEqual(
            new_paths,
            [
                "s3://bucket/common-crawl/attributes/cc_head",
                "s3://bucket/common-crawl/attributes/cc_middle",
                "s3://bucket/common-crawl/attributes/cc_tail",
            ],
        )

        paths = [
            "s3://bucket/common-crawl/documents/*.json.gz",
            "s3://bucket2/c4/documents/**/data/*.json.gz",
        ]
        new_paths = _make_paths_from_prefix(
            paths=paths,
            prefix="/local/path/",
        )
        self.assertEqual(
            new_paths,
            [
                "/local/path/bucket/common-crawl/documents",
                "/local/path/bucket2/c4/documents",
            ],
        )

    def test_runtime_e2e(self):
        documents_path = f"{LOCAL_DATA}/provided/documents/000.json.gz"
        experiment_name = "test"
        taggers = ["c4_v1"]

        with TemporaryDirectory() as temp_dir:
            create_and_run_tagger(
                documents=[documents_path],
                destination=temp_dir,
                taggers=taggers,
                experiment=experiment_name,
                debug=True,
            )
            destination_file = os.path.join(temp_dir, experiment_name, "000.json.gz")
            self.assertTrue(os.path.exists(destination_file))

            with smart_open.open(documents_path, "rt") as f:
                document = [json.loads(ln) for ln in f]

            with smart_open.open(destination_file, "rt") as f:
                attributes = [json.loads(ln) for ln in f]

        self.assertEqual(len(document), len(attributes))

        for d, a in zip(document, attributes):
            self.assertEqual(d["id"], a["id"])
            self.assertTrue(sorted(a.keys()), ["attributes", "id", "source"])

            for key, value in a["attributes"].items():
                parts = key.split("__")
                self.assertEqual(len(parts), 3)
                self.assertEqual(parts[0], experiment_name)
                self.assertTrue(parts[1] in taggers)

                for elem in value:
                    self.assertEqual(len(elem), 3)
                    self.assertTrue(isinstance(elem[0], int))
                    self.assertTrue(isinstance(elem[1], int))
                    self.assertTrue(isinstance(elem[2], float))

                if len(value) == 1:
                    self.assertEqual(value[0][0], 0)
                    self.assertEqual(value[0][1], len(d["text"]))

    def test_alt_src(self):
        taggers = ["c4_v1"]
        experiment_name = "test"

        with TemporaryDirectory() as temp_dir:
            create_and_run_tagger(
                documents=[f"{LOCAL_DATA}/provided/documents/000.json.gz"],
                destination=temp_dir,
                taggers=taggers,
                experiment=experiment_name,
                debug=True,
            )
            destination_file = os.path.join(temp_dir, experiment_name, "000.json.gz")
            self.assertTrue(os.path.exists(destination_file))

            with smart_open.open(destination_file, "rt") as f:
                attributes_full_name = [json.loads(ln) for ln in f]

        with TemporaryDirectory() as temp_dir:
            create_and_run_tagger(
                documents=[f"{LOCAL_DATA}/provided/documents/*"],
                destination=temp_dir,
                taggers=taggers,
                experiment=experiment_name,
                debug=True,
            )
            destination_file = os.path.join(temp_dir, experiment_name, "000.json.gz")
            self.assertTrue(os.path.exists(destination_file))

            with smart_open.open(destination_file, "rt") as f:
                attributes_star_in_path = [json.loads(ln) for ln in f]

        with TemporaryDirectory() as temp_dir:
            create_and_run_tagger(
                documents=[f"{LOCAL_DATA}/provided/documents/"],
                destination=temp_dir,
                taggers=taggers,
                experiment=experiment_name,
                debug=True,
            )
            destination_file = os.path.join(temp_dir, experiment_name, "000.json.gz")
            self.assertTrue(os.path.exists(destination_file))

            with smart_open.open(destination_file, "rt") as f:
                attributes_only_dir = [json.loads(ln) for ln in f]

        self.assertEqual(attributes_full_name, attributes_star_in_path)
        self.assertEqual(attributes_full_name, attributes_only_dir)

    def test_alt_exp(self):
        documents_path = f"{LOCAL_DATA}/provided/documents/000.json.gz"
        taggers = ["c4_v1"]

        with TemporaryDirectory() as temp_dir:
            create_and_run_tagger(
                documents=[documents_path],
                destination=temp_dir,
                taggers=taggers,
                experiment="test",
                debug=True,
            )
            destination_file = os.path.join(temp_dir, "test", "000.json.gz")
            self.assertTrue(os.path.exists(destination_file))

            with smart_open.open(destination_file, "rt") as f:
                attributes_with_exp = [json.loads(ln) for ln in f]

        with TemporaryDirectory() as temp_dir:
            create_and_run_tagger(
                documents=[documents_path],
                destination=temp_dir,
                taggers=taggers,
                experiment=None,
                debug=True,
            )
            destination_file = os.path.join(temp_dir, "c4_v1", "000.json.gz")
            self.assertTrue(os.path.exists(destination_file))

            with smart_open.open(destination_file, "rt") as f:
                attributes_without_exp = [json.loads(ln) for ln in f]

        for row_with_exp, row_without_exp in zip(attributes_with_exp, attributes_without_exp):
            for key_with_exp, key_without_exp in zip(row_with_exp["attributes"], row_without_exp["attributes"]):
                parts_with_exp = key_with_exp.split("__")
                parts_without_exp = key_without_exp.split("__")
                self.assertNotEqual(parts_with_exp[0], parts_without_exp[0])
                self.assertEqual(parts_with_exp[1], parts_without_exp[1])
                self.assertEqual(parts_with_exp[2], parts_without_exp[2])
                self.assertEqual(parts_with_exp[0], "test")
                self.assertEqual(parts_without_exp[0], "c4_v1")



import multiprocessing
from argparse import ArgumentParser
from pathlib import Path
from typing import List, Optional

from yaml import safe_load

from .analyzer import AnalyzerCli
from .deduper import DeduperCli
from .mixer import MixerCli

# must import these to register the resolvers
from .resolvers import *  # noqa: F401,F403,W0401
from .tagger import ListTaggerCli, TaggerCli
from .tokenizer import TokenizerCli

AVAILABLE_COMMANDS = {
    "dedupe": DeduperCli,
    "mix": MixerCli,
    "tag": TaggerCli,
    "list": ListTaggerCli,
    "stat": AnalyzerCli,
    "tokens": TokenizerCli,
    # following functionality is not yet implemented
    # "train-ft": None,
    # "train-lm": None,
}


def main(argv: Optional[List[str]] = None):
    try:
        # attempting to set start method to spawn in case it is not set
        multiprocessing.set_start_method("spawn")
    except RuntimeError:
        # method already set, check if it is set to spawn
        if multiprocessing.get_start_method() != "spawn":
            raise RuntimeError("Multiprocessing start method must be set to spawn")

    parser = ArgumentParser(
        prog="dolma",
        usage="dolma [command] [options]",
        description="Command line interface for the DOLMa dataset processing toolkit",
    )
    parser.add_argument(
        "-c",
        "--config",
        help="Path to configuration optional file",
        type=Path,
        default=None,
    )
    subparsers = parser.add_subparsers(dest="command")
    subparsers.required = True
    subparsers.choices = AVAILABLE_COMMANDS.keys()  # type: ignore

    for command, cli in AVAILABLE_COMMANDS.items():
        cli.make_parser(subparsers.add_parser(command, help=cli.DESCRIPTION))

    args = parser.parse_args(argv)

    # try parsing the config file
    config: Optional[dict] = None
    if config_path := args.__dict__.pop("config"):
        assert config_path.exists(), f"Config file {config_path} does not exist"
        with open(config_path) as f:
            config = dict(safe_load(f))

    AVAILABLE_COMMANDS[args.__dict__.pop("command")].run_from_args(args=args, config=config)



"""
Copied from https://github.com/bigcode-project/bigcode-dataset/blob/main/preprocessing/utils/text_extraction.py

Extract Python comments (using Python tokenizer) and
docstrings (using AST parsing).
Extract Java and JavaScript comments (using Pygments)
"""

import ast
import io
import tokenize
import warnings
from itertools import groupby
from os.path import basename, splitext

import pygments
from pygments.lexers import get_lexer_by_name
from pygments.token import Comment

StringIO = io.StringIO

NODE_TYPES = {
    ast.ClassDef: "Class",
    ast.FunctionDef: "Function/Method",
    ast.Module: "Module",
}


# comment extraction
def get_comments(s, clean=False):
    "Returns a string including all comments in python code"
    coments = []
    g = tokenize.generate_tokens(StringIO(s).readline)
    for toknum, tokval, _, _, _ in g:
        # print(toknum,tokval)
        if toknum == tokenize.COMMENT:
            coments.append((toknum, tokval))
    result = tokenize.untokenize(coments)
    if clean:
        result = result.replace("#", "")
    return result


# Note: sometimes this can miss examples with decorators over classes
# ast parsing, source: https://gist.github.com/SpotlightKid/1548cb6c97f2a844f72d
def parse_docstrings(source):
    """Parse Python source code and yield a tuple of ast node instance, name,
    and docstring for each function/method, class and module."""
    tree = ast.parse(source)

    for node in ast.walk(tree):
        if isinstance(node, tuple(NODE_TYPES)):
            docstring = ast.get_docstring(node)  # pyright: ignore

            yield (node, getattr(node, "name", None), docstring)


def get_docstrings(source, module="<string>"):
    """Parse Python source code from file or string and print docstrings."""
    if hasattr(source, "read"):
        filename = getattr(source, "name", module)
        module = splitext(basename(filename))[0]
        source = source.read()

    docstrings = sorted(
        parse_docstrings(source), key=lambda x: (NODE_TYPES.get(type(x[0])), x[1])  # pyright: ignore
    )

    grouped = groupby(docstrings, key=lambda x: NODE_TYPES.get(type(x[0])))  # pyright: ignore
    results = []
    for _, group in grouped:
        for _, name, docstring in group:
            name = name if name else module
            # print(docstring or '')
            if docstring:
                results.append(docstring)
    return results


def get_text_python(source, comments=True, clean_comments=True):
    """Extract all natural text in source: comments + docstrings
    the extraction fails in case of syntax errors in the file
    Args:
        source: the code to parse
        comments: if True extract comments two
        clean_comment: if True remove # from extracted comments
    Returns:
        a string with concatenated docstrings and comments"""

    try:
        docstrings = "\n".join(get_docstrings(source))
    except:  # pylint: disable=bare-except   # noqa: E722
        docstrings = ""
        warnings.warn("code couldn't be parsed due to compilation failure, no docstring is extracted")

    if comments:
        try:
            comments = get_comments(source, clean=clean_comments)
        except:  # pylint: disable=bare-except    # noqa: E722
            comments = ""
            warnings.warn("tokenization error, no comment is extracted")
    else:
        comments = ""

    output = docstrings + "\n" + comments
    return output.strip()


def comment_size(text, language):
    """
    Compute the size of comments in a program (not necessarily python).
    """
    lexer = get_lexer_by_name(language)
    tokens = pygments.lex(text, lexer)
    comment_len = 0
    for token_type, token in tokens:
        if token_type == Comment.Multiline or token_type == Comment.Single:
            comment_len += len(token)  # token is a string with the comment contents
    return comment_len


def get_nl_ratio(text, language):
    """get the ratio of comments to code in a program"""
    if language == "python":
        comments = get_text_python(text)
        ratio = len(comments) / len(text)
    else:
        ratio = comment_size(text, language) / len(text)
    return ratio



import argparse

from .executor import tokenize_in_parallel


def parse_args() -> argparse.Namespace:
    ap = argparse.ArgumentParser()
    ap.add_argument("--sources", nargs="+", required=True)
    ap.add_argument("--destination", required=True)
    ap.add_argument("--tokenizer-id", type=str, default="allenai/eleuther-ai-gpt-neox-20b-pii-special")
    ap.add_argument("--metadata-dir", type=str, default=None)
    ap.add_argument("--num-tokenizers", type=int, default=1)
    ap.add_argument("--num-writers", type=int, default=1)
    ap.add_argument("--max-size", type=int, default=1024 * 1024 * 1024)
    return ap.parse_args()


def main():
    args = parse_args()
    tokenize_in_parallel(
        sources=args.sources,
        destination=args.destination,
        tokenizer_name_or_path=args.tokenizer_id,
        metadata_dir=args.metadata_dir,
        num_tokenizers=args.num_tokenizers,
        num_writers=args.num_writers,
        max_size=args.max_size,
    )


if __name__ == "__main__":
    main()



"""
Minimal example to download a Wikipedia dump, process via WikiExtractor, and transform to Dolma format.

Author: Luca Soldaini (@soldni)
"""


import argparse
import datetime
import json
import logging
import multiprocessing
import re
import string
import sys
from contextlib import ExitStack
from io import TextIOWrapper
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Dict, List, Union

import requests
import smart_open
from uniseg.wordbreak import words as uniseg_get_words

from dolma.core.parallel import BaseParallelProcessor, QueueType

CMD_INSTALL = "pip install git+https://github.com/santhoshtr/wikiextractor.git requests smart_open tqdm"

try:
    from wikiextractor import WikiExtractor
except ImportError:
    print(f"Please install wikiextractor with `{CMD_INSTALL}`")
    sys.exit(1)

try:
    import requests  # noqa
except ImportError:
    print(f"Please install requests with `{CMD_INSTALL}`")
    sys.exit(1)

try:
    import smart_open  # noqa
except ImportError:
    print(f"Please install smart_open with `{CMD_INSTALL}`")
    sys.exit(1)

try:
    import tqdm  # noqa
except ImportError:
    print(f"Please install tqdm with `{CMD_INSTALL}`")
    sys.exit(1)


DUMP_URL = "https://dumps.wikimedia.org/{lang}wiki/{date}/{lang}wiki-{date}-pages-articles-multistream.xml.bz2"
LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


def convert_timestamp(d: datetime.datetime) -> str:
    return d.strftime("%Y-%m-%dT%H:%M:%S.%f")[:23] + "Z"


def get_words(text: str) -> List[str]:
    return [word for word in uniseg_get_words(text) if not all(char in string.whitespace for char in word)]


def download_file(url, filename):
    response = requests.get(url, stream=True)

    if response.status_code != 200:
        raise ValueError(f"File {url} not found; try a more recent date")

    file_size = int(response.headers["Content-Length"])

    with open(filename, "wb") as f, tqdm.tqdm(
        desc="Downloading", total=file_size, unit="B", unit_scale=True, unit_divisor=1024
    ) as pbar:
        for chunk in response.iter_content(chunk_size=1024 * 1024):
            if chunk:  # filter out keep-alive new chunks
                f.write(chunk)
                pbar.update(len(chunk))


def download_wiki(date: str, lang: str, output_path: Union[str, Path], overwrite: bool = False):
    assert re.match(r"\d{8}", date), "Date must be in YYYYMMDD format"
    dump_url = DUMP_URL.format(date=date, lang=lang)

    output_path = Path(str(output_path))
    output_path.parent.mkdir(parents=True, exist_ok=True)

    if output_path.exists() and not overwrite:
        LOGGER.warning(f"Output file {output_path} exists. Use --overwrite to overwrite it.")
        return

    if not output_path.suffix.endswith(".bz2"):
        LOGGER.warning("Output file does not end with .bz2. This is not recommended.")

    try:
        LOGGER.info(f"Downloading {dump_url} to {output_path}")
        download_file(dump_url, output_path)
    except Exception as exception:
        if output_path.exists():
            output_path.unlink()
        raise exception


def get_arguments():
    ap = argparse.ArgumentParser()
    ap.add_argument("--date", help="Date in YYYYMMDD format", default="20231001")
    ap.add_argument("--lang", help="Language code", default="simple")
    ap.add_argument("--output", help="Output dir", required=True)
    ap.add_argument("--overwrite", help="Overwrite existing file", action="store_true")
    ap.add_argument("--processes", help="Number of processes", type=int, default=1)
    return ap.parse_args()


class NewOutputSplitter:
    file: TextIOWrapper

    def __init__(self, nextFile, max_file_size=0, compress=True):
        """
        :param nextFile: a NextFile object from which to obtain filenames
            to use.
        :param max_file_size: the maximum size of each file.
        :para compress: whether to write data with bzip compression.
        """
        self.nextFile = nextFile
        self.compress = compress
        self.max_file_size = max_file_size
        self.stack = ExitStack()
        self.ext = ".gz" if compress else ""
        self.written = 0
        self.open()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def reserve(self, size):
        if self.written + size > self.max_file_size:
            self.close()
            self.open()
            self.written = 0

    def write(self, data):
        self.reserve(len(data))
        self.file.write(data)

    def close(self):
        self.stack.pop_all().close()

    def open(self):
        self.file = self.stack.enter_context(smart_open.open(self.nextFile.next() + self.ext, "wt"))


def wiki_extract(output_gzip: Path, output_json: Path, processes: int = 1, overwrite: bool = False):
    if output_json.exists() and not overwrite:
        LOGGER.warning(f"Output file {output_json} exists. Use --overwrite to overwrite it.")
        return

    WikiExtractor.Extractor.keepLinks = False
    WikiExtractor.Extractor.HtmlFormatting = False
    WikiExtractor.ignoreTag("a")
    WikiExtractor.Extractor.to_json = True  # type: ignore
    WikiExtractor.acceptedNamespaces = ["Article"]
    # input_file = output_gzip

    # # output_path = output_json
    # def _open(*args, **kwargs):
    #     return tqdm.tqdm(smart_open.open(*args, **kwargs))

    WikiExtractor.decode_open = smart_open.open
    WikiExtractor.OutputSplitter = NewOutputSplitter

    file_size = 1024 * 1024 * 1024  # 1 GB

    output_json.mkdir(exist_ok=True, parents=True)

    WikiExtractor.process_dump(
        input_file=str(output_gzip),
        template_file=None,
        out_file=str(output_json),
        file_size=file_size,
        file_compress=True,
        process_count=processes,
        html_safe=False,
    )


class WikiExtractorParallel(BaseParallelProcessor):
    @classmethod
    def increment_progressbar(  # type: ignore[override]
        cls, queue, /, files: int = 0, documents: int = 0
    ) -> Dict[str, int]:
        return super().increment_progressbar(
            queue,
            files=files,
            documents=documents,
        )

    @classmethod
    def process_single(cls, source_path: str, destination_path: str, queue: QueueType, **kwargs: Any):
        date = kwargs.pop("date", None)
        lang = kwargs.pop("lang", None)
        assert date is not None, "date should be provided"
        assert lang is not None, "lang should be provided"

        created_time = convert_timestamp(datetime.datetime.strptime(date, "%Y%m%d"))
        current_time = convert_timestamp(datetime.datetime.now())
        logger = cls.get_logger()

        documents_count = 0
        update_interval = 1

        with smart_open.open(source_path) as f, smart_open.open(destination_path, "w") as g:
            try:
                for i, line in enumerate(f):
                    try:
                        data = json.loads(line)
                    except json.JSONDecodeError as exp:
                        logger.warning("Failed to parse %s:%s `%s...`: %s", source_path, i, line[:80], exp)
                        continue

                    id_ = data.pop("id")
                    title = data.pop("title", "").strip()
                    body = data.pop("text", "").strip()

                    if not id_ or not title or not body:
                        # logger.warning("Skipping %s:%s: missing id, title or body", source_path, i)
                        continue

                    text = f"{title}\n\n{body}".strip()

                    json_data = {
                        "id": id_,
                        "source": "wikipedia",
                        "version": "v0",
                        "text": text,
                        "created": created_time,
                        "added": current_time,
                        "metadata": {**data, "length": len(get_words(text))},
                    }

                    g.write(json.dumps(json_data) + "\n")
                    documents_count += 1

                    if documents_count % update_interval == 0:
                        # update the progress bar every 1000 documents to prevent
                        # buffering
                        cls.increment_progressbar(queue, documents=documents_count)

                        if queue.qsize() >= multiprocessing.cpu_count():
                            # double the update interval if the queue is full
                            update_interval *= 2

                        documents_count = 0

            except Exception as exp:
                logger.warning("Failed to process %s: %s", source_path, exp)
                return

            cls.increment_progressbar(queue, files=1, documents=documents_count)


def main():
    args = get_arguments()
    output_gzip = Path(args.output) / f"wiki_{args.date}_{args.lang}.xml.bz2"
    output_json = Path(args.output) / f"wiki_{args.date}_{args.lang}"
    output_final = Path(args.output) / "v0/documents"

    download_wiki(date=args.date, lang=args.lang, output_path=output_gzip, overwrite=args.overwrite)
    wiki_extract(output_gzip=output_gzip, output_json=output_json, processes=args.processes)

    with TemporaryDirectory() as tempdir:
        processor = WikiExtractorParallel(
            source_prefix=f"{output_json}/*/*.gz",
            destination_prefix=str(output_final),
            metadata_prefix=tempdir,
            num_processes=args.processes,
        )
        processor(date=args.date, lang=args.lang)


if __name__ == "__main__":
    # setting multiprocessing start method to spawn
    multiprocessing.set_start_method("spawn")

    main()



"""
Utilities to work with a OmegaConf structured config object

Author: Luca Soldaini (@soldni)
"""


from argparse import ArgumentParser, Namespace
from collections.abc import Iterable
from copy import deepcopy
from dataclasses import Field
from dataclasses import field as dataclass_field
from dataclasses import is_dataclass
from logging import warn
from typing import (
    Any,
    Dict,
    Generic,
    Literal,
    Optional,
    Protocol,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
)

from omegaconf import MISSING, DictConfig, ListConfig
from omegaconf import OmegaConf as om
from omegaconf.errors import OmegaConfBaseException
from rich.console import Console
from rich.syntax import Syntax

from dolma.core.errors import DolmaConfigError

__all__ = [
    "BaseCli",
    "field",
    "make_parser",
    "namespace_to_nested_omegaconf",
    "print_config",
]


T = TypeVar("T", bound=Any)
D = TypeVar("D", bound="DataClass")
A = TypeVar("A", bound="ArgumentParser")


def _field_nargs(default: Any) -> Union[Literal["?"], Literal["*"]]:
    # return '+' if _default is iterable but not string/bytes, else 1
    if isinstance(default, (str, bytes)):
        return "?"

    if isinstance(default, Iterable):
        return "*"

    return "?"


def field(default: T = MISSING, help: Optional[str] = None, **extra: Any) -> T:
    metadata = {"help": help, "type": type(default), "default": default, "nargs": _field_nargs(default), **extra}
    return dataclass_field(default_factory=lambda: deepcopy(default), metadata=metadata)


class DataClass(Protocol):
    __dataclass_fields__: Dict[str, Field]


def make_parser(parser: A, config: Type[DataClass], prefix: Optional[str] = None) -> A:
    for field_name, dt_field in config.__dataclass_fields__.items():
        # get type from annotations or metadata
        typ_ = config.__annotations__.get(field_name, dt_field.metadata.get("type", MISSING))

        if typ_ is MISSING:
            warn(f"No type annotation for field {field_name} in {config.__name__}")
            continue

        # join prefix and field name
        field_name = f"{prefix}.{field_name}" if prefix else field_name

        # This section here is to handle Optional[T] types
        # We only care for cases where T is a dataclass
        # So we first check if type is Union since Optional[T] is just a shorthand for Union[T, None]
        # and that the union contains only one non-None type
        if get_origin(typ_) == Union:
            # get all non-None types
            args = [a for a in get_args(typ_) if a is not type(None)]  # noqa: E721

            if len(args) == 1:
                # simple Optional[T] type
                typ_ = args[0]

        # here's where we check if T is a dataclass
        if is_dataclass(typ_):
            # recursively add subparsers
            make_parser(parser, typ_, prefix=field_name)
            continue

        if typ_ is bool:
            # for boolean values, we add two arguments: --field_name and --no-field_name
            parser.add_argument(
                f"--{field_name}",
                help=dt_field.metadata.get("help"),
                dest=field_name,
                action="store_true",
                default=MISSING,
            )
            parser.add_argument(
                f"--no-{field_name}",
                help=f"Disable {field_name}",
                dest=field_name,
                action="store_false",
                default=MISSING,
            )
        else:
            # else it's just a normal argument
            parser.add_argument(
                f"--{field_name}",
                help=dt_field.metadata.get("help"),
                nargs=dt_field.metadata.get("nargs", "?"),
                default=MISSING,
            )

    return parser


def _make_nested_dict(key: str, value: Any, d: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
    d = d or {}

    if "." in key:
        key, rest = key.split(".", 1)
        value = _make_nested_dict(rest, value, d.get(key))

    if value is not MISSING:
        d[key] = value

    return d


def namespace_to_nested_omegaconf(args: Namespace, structured: Type[T], config: Optional[dict] = None) -> T:
    nested_config_dict: Dict[str, Any] = {}
    for key, value in vars(args).items():
        nested_config_dict = _make_nested_dict(key, value, nested_config_dict)

    untyped_config: DictConfig = om.merge(
        om.create(config or {}), om.create(nested_config_dict)
    )  # pyright: ignore (pylance is confused because om.create might return a DictConfig or a ListConfig)

    base_structured_config: DictConfig = om.structured(structured)
    merged_config = om.merge(base_structured_config, untyped_config)
    assert isinstance(merged_config, DictConfig)
    return merged_config  # pyright: ignore


def print_config(config: Any, console: Optional[Console] = None) -> None:
    if not isinstance(config, (DictConfig, ListConfig)):
        config = om.create(config)

    # print the config as yaml using a rich syntax highlighter
    console = console or Console()
    yaml_config = om.to_yaml(config, sort_keys=True).strip()
    highlighted = Syntax(code=yaml_config, lexer="yaml", theme="ansi_dark")
    console.print(highlighted)


class BaseCli(Generic[D]):
    CONFIG: Type[D]
    DESCRIPTION: Optional[str] = None

    @classmethod
    def make_parser(cls, parser: A) -> A:
        assert hasattr(cls, "CONFIG"), f"{cls.__name__} must have a CONFIG attribute"
        return make_parser(parser, cls.CONFIG)

    @classmethod
    def run_from_args(cls, args: Namespace, config: Optional[dict] = None):
        assert hasattr(cls, "CONFIG"), f"{cls.__name__} must have a CONFIG attribute"
        parsed_config = namespace_to_nested_omegaconf(args=args, structured=cls.CONFIG, config=config)
        try:
            return cls.run(parsed_config)
        except OmegaConfBaseException as ex:
            raise DolmaConfigError(
                f"Invalid error while parsing key `{ex.full_key}` of `{ex.object_type_str}`: "
                f"{type(ex).__name__}"
            ) from ex

    @classmethod
    def run(cls, parsed_config: D):
        raise NotImplementedError("Abstract method; must be implemented in subclass")



"""

Utilities for code-related taggers.

@akshitab, @soldni

"""
import json
import logging
from pathlib import Path
from typing import Dict, Generator

import regex
import smart_open
from bs4 import BeautifulSoup  # pylint: disable=import-error
from detect_secrets.core.potential_secret import PotentialSecret
from detect_secrets.core.scan import _process_line_based_plugins
from detect_secrets.core.secrets_collection import SecretsCollection
from detect_secrets.settings import default_settings, get_plugins

logger = logging.getLogger(__name__)


def scan_code(code: str) -> Generator["PotentialSecret", None, None]:
    if not get_plugins():
        logger.error("No plugins to scan with!")
        return

    has_secret = False
    for lines in [code.splitlines()]:
        for secret in _process_line_based_plugins(
            lines=list(enumerate(lines, start=1)),
            filename="code_str.yml",
        ):
            has_secret = True
            yield secret

        if has_secret:
            break


class SecretsCollectionForStringInput(SecretsCollection):
    def scan_str(self, code_str: str):
        for secret in scan_code(code_str):
            self["code_str.yml"].add(secret)


def get_secrets(code: str):
    secrets = SecretsCollectionForStringInput()
    with default_settings():
        secrets.scan_str(code)

    return secrets


def filter_html(html: str) -> float:
    """Filter HTML files based on displayed text VS code ratio"""
    try:
        soup = BeautifulSoup(html, features="html.parser")
    except (TypeError, UnboundLocalError):
        return False

    # kill all script and style elements
    for script in soup(["script", "style"]):
        script.extract()  # rip it out

    # get text
    text = soup.get_text()
    ratio = len(text) / len(html)

    return (ratio) * (len(text) > 100)


def get_whitespace_regex() -> regex.Pattern:
    return regex.compile(r"\w+|[^\w\s]+")


def get_ext_to_lang_mapping() -> Dict[str, str]:
    path = Path(__file__).parent / "../../data/ext_to_lang_mapping.json"
    with smart_open.open(path, "r") as f:
        return json.load(f)



"""

Filters.

@kylel, @soldni

"""

from typing import Iterable

from ..core.data_types import TextSlice
from ..core.ft_tagger import BaseFastTextTagger, Prediction
from ..core.registry import TaggerRegistry


@TaggerRegistry.add("jigsaw_hatespeech_document_v2")
class FastTextJigsawHatespeechDocumentTagger(BaseFastTextTagger):
    MODEL_PATH = "https://ai2-s2-research-public.s3.us-west-2.amazonaws.com/aakankshan/olmo-data-filters/jigsaw_fasttext_bigrams_hatespeech_final.bin"  # noqa: E501

    def __init__(self):
        super().__init__(model_path=self.MODEL_PATH, model_mode=self.DOCUMENT_LEVEL_TAGGER)

    def predict_slice(self, text_slice: TextSlice) -> Iterable[Prediction]:
        labels, probs = self.classifier.predict(text_slice.text.replace("\n", " ").strip(), k=-1)
        label_index = 1 if "non" in labels[0] else 0  # pyright: ignore
        return (
            Prediction(label=labels[label_index], score=probs[label_index]),
            Prediction(label=labels[1 - label_index], score=probs[1 - label_index]),
        )


@TaggerRegistry.add("jigsaw_hatespeech_sentence_v2")
class FastTextJigsawHatespeechSentenceTagger(FastTextJigsawHatespeechDocumentTagger):
    def __init__(self):
        BaseFastTextTagger.__init__(self, model_path=self.MODEL_PATH, model_mode=self.SENTENCE_LEVEL_TAGGER)


@TaggerRegistry.add("jigsaw_nsfw_document_v1")
class FastTextJigsawNsfwDocumentTagger(FastTextJigsawHatespeechDocumentTagger):
    MODEL_PATH = "https://ai2-s2-research-public.s3.us-west-2.amazonaws.com/aakankshan/olmo-data-filters/jigsaw_fasttext_bigrams_nsfw_final.bin"  # noqa: E501


@TaggerRegistry.add("jigsaw_nsfw_sencence_v2")
class FastTextJigsawNsfwSentenceTagger(FastTextJigsawHatespeechSentenceTagger):
    MODEL_PATH = "https://ai2-s2-research-public.s3.us-west-2.amazonaws.com/aakankshan/olmo-data-filters/jigsaw_fasttext_bigrams_nsfw_final.bin"  # noqa: E501



"""
Evaluating reasoning abilities of model ablations with varying amount of code in pre-training
"""
import abc
import argparse
import os
import random

import numpy as np
import pandas as pd
import torch

from abc import ABC

from torch.utils.data import DataLoader
from accelerate import Accelerator
from tqdm import tqdm

from datasets import load_dataset
from hf_olmo import OLMoForCausalLM, OLMoTokenizerFast
from rouge_score import rouge_scorer
from transformers import DataCollatorForSeq2Seq, get_scheduler

models_checkpoints = {
    "c4-stack-5p": "../checkpoints/c4-stack-5p/step60000-unsharded/",
    "c4-stack-15p": "../checkpoints/c4-stack-15p/step60000-unsharded/",
    "GPT-Neox-20B-0p": "../checkpoints/GPT-Neox-20B-0p/step60000-unsharded/",
}


def get_args():
    args = argparse.ArgumentParser()

    args.add_argument("dataset", type=str, choices=["babi", "web_nlg", "gsm8k"])
    args.add_argument("--models", nargs="+", default=None)
    args.add_argument("--n-test-samples", type=int, default=None)
    args.add_argument("--n-train-samples", type=int, default=None)
    args.add_argument("--seed", type=int, default=0)
    args.add_argument("--max-few-shots", type=int, default=5)
    args.add_argument("--number-seeds", type=int, default=5)

    # only used for GSM8K train/eval
    args.add_argument("--per_device_train_batch_size", type=int, default=8, help="Batch size")
    args.add_argument("--test_per_device_train_batch_size", type=int, default=16, help="Test Batch size")
    args.add_argument("--learning_rate", type=float, default=5e-5, help="Learning rate")
    args.add_argument("--n-solutions-sampled", default=20, type=int, help="Number of solutions to sample per example")
    args.add_argument("--num-train-epochs", type=int, default=3, help="Number of training epochs")
    args.add_argument("--warmup_ratio", type=float, default=0.03, help="Ratio of total training steps used for warmup.")

    return args.parse_args()


class Eval(ABC):
    @abc.abstractmethod
    def build_prompt(self, test_example, demonstrations):
        pass

    @abc.abstractmethod
    def get_num_new_tokens(self, test_example, tokenizer):
        pass

    @abc.abstractmethod
    def compute_metrics(self, prediction, test_example):
        pass


class BabiEval(Eval):
    def build_prompt(self, test_example, demonstrations):
        prompt = ""
        for demonstration in demonstrations:
            prompt += demonstration["passage"] + demonstration["question"] + " " + demonstration["answer"] + "\n\n"

        prompt += test_example["passage"] + test_example["question"]

        return prompt

    def get_num_new_tokens(self, test_example, tokenizer):
        # generate a bit more tokens than gold (otherwise we might think prediction is correct even though the actual prediction is longer, thus wrong) - this can be replaced with a stop token
        return len(tokenizer.encode(test_example["answer"], add_special_tokens=False)) + 5

    def compute_metrics(self, prediction, test_example):
        return {"accuracy": int(prediction == test_example["answer"])}


class WebNLGEval(Eval):
    def __init__(self):
        self.scorer = rouge_scorer.RougeScorer(['rouge2'], use_stemmer=True)

    def build_prompt(self, test_example, demonstrations):
        prompt = """I will verbalize an abstract representation of a sentence in natural language. To do so, I will first show the representation and then the natural language. The text needs to include all of the information in the representation.\n\n"""
        for demonstration in demonstrations:
            prompt += ", ".join(demonstration["input"]) + "\n"
            prompt += f"{demonstration['target']}\n\n"

        prompt += ", ".join(test_example["input"]) + "\n"

        return prompt

    def get_num_new_tokens(self, test_example, tokenizer):
        return 200  # this is enough for the test set

    def compute_metrics(self, prediction, test_example):
        return {"rouge2_f1": self.scorer.score(test_example["target"], prediction)['rouge2'].fmeasure}

class GSM8KEval(Eval):
    def build_prompt(self, test_example, demonstrations):
        return ""

    def get_num_new_tokens(self, test_example, tokenizer):
        return 300  # this is enough for the test set

    def compute_metrics(self, prediction, test_example):

        def run_program(code):
            """Important: executing code outside a secure docker container is potentially dangerous"""
            if "import" in code:
                raise ValueError("Import is not allowed")
            if "open(" in code:
                raise ValueError("Open is not allowed")
            if "eval(" in code:
                raise ValueError("Eval is not allowed")
            if "exec(" in code:
                raise ValueError("Exec is not allowed")
            if "compile(" in code:
                raise ValueError("Compile is not allowed")

            try:
                exec(code)
                answer = locals()["solution"]()
                return {
                    "success": True,
                    "answer": answer,
                    "error": None,
                }
            except Exception as e:
                return {
                    "success": False,
                    "answer": None,
                    "error": str(e),
                }

        gold_answer = test_example["answer"].split("####")[1].strip()

        try:
            predicted_answer = run_program(prediction)["answer"]
            return {"accuracy": int(float(predicted_answer) == float(gold_answer))}
        except Exception:
            return {"accuracy": 0}


def complete_prompt(model, tokenizer, prompt: str, new_tokens=100):
    input_ids = tokenizer.encode(prompt, add_special_tokens=False)
    max_length = len(input_ids) + new_tokens

    input_ids = torch.Tensor([input_ids]).long().to(model.device)
    attention_mask = torch.ones_like(input_ids).to(model.device)
    entire_output = model.generate(input_ids, attention_mask=attention_mask, max_length=max_length)[0]
    output_new_tokens = entire_output[len(input_ids[0]):]
    decoded = tokenizer.decode(output_new_tokens, skip_special_tokens=True).strip()

    return decoded


def eval_only(evaluator, model, model_name, tokenizer, train_dataset, test_dataset, seed, args):
    """
    Evaluate the model on the test set by averaging over multiple number of few-shot examples and seeds, similar setup to https://arxiv.org/abs/2305.16264
    """
    train_dataset = list(train_dataset)
    test_dataset = list(test_dataset)

    results = []
    agg_results = []

    for few_shot in range(args.max_few_shots + 1):
        tqdm_loop = tqdm(test_dataset)

        all_metrics = []

        for ex in tqdm_loop:
            # always pick the same demonstrations for a given seed
            random.seed(seed)

            demonstrations = random.sample(train_dataset, few_shot)

            try:
                context = evaluator.build_prompt(ex, demonstrations)
                n_new_tokens = evaluator.get_num_new_tokens(ex, tokenizer)

                prediction = complete_prompt(model=model, tokenizer=tokenizer, prompt=context,
                                             new_tokens=n_new_tokens)
                prediction = prediction.split("\n")[0].strip()

                ex_metrics = evaluator.compute_metrics(prediction, ex)

                all_metrics.append(ex_metrics)
                avg_metrics = {key: np.mean([ex[key] for ex in all_metrics]) for key in all_metrics[0].keys()}

                results.append({
                    "model": model_name,
                    "few_shot": few_shot,
                    "seed": seed,
                    "context": context,
                    "prediction": prediction,
                    "answer": ex.get("answer") or ex.get("target"),
                    **ex_metrics
                })
                tqdm_loop.set_description(
                    f"model: {model_name}, few_shots: {few_shot}, seed: {seed}, " + ", ".join(
                        [f"{key}: {value:.3f}" for key, value in avg_metrics.items()]))
            except Exception as e:
                print(e)
                print(f"Skipping...")
                continue

        avg_metrics = {key: np.mean([ex[key] for ex in all_metrics]) for key in all_metrics[0].keys()}
        agg_results.append({
            "model": model_name,
            "seed": seed,
            "few_shot": few_shot,
            **avg_metrics
        })

    return results, agg_results


def train_and_eval_gsm8k(evaluator, model, model_name, tokenizer, test_dataset, seed, args):
    def tokenize_function_train(example):
        output_ids = tokenizer.encode(example["answer"], add_special_tokens=False) + [tokenizer.eos_token_id]
        input_ids = tokenizer.encode(example["question"], add_special_tokens=False) + output_ids
        labels = [-100] * len(tokenizer.encode(example["question"], add_special_tokens=False)) + output_ids
        return {"input_ids": input_ids, "labels": labels}

    def tokenize_function_eval(example):
        input_ids = tokenizer.encode(example["question"], add_special_tokens=False)
        labels = tokenizer.encode(example["answer"], add_special_tokens=False) + [tokenizer.eos_token_id]
        return {"input_ids": input_ids, "labels": labels}

    def evaluate(model, tokenizer, dataset, dataset_loader):
        eval_result = []
        eval_loop = tqdm(dataset_loader, desc="Evaluating")
        accuracies = []
        examples = list(dataset)
        i = 0
        for eval_batch in eval_loop:
            max_length = eval_batch["input_ids"].size(1) + evaluator.get_num_new_tokens(None, tokenizer)
            entire_output = model.generate(eval_batch["input_ids"], attention_mask=eval_batch["attention_mask"],
                                           max_length=max_length, do_sample=True, temperature=0.7, top_p=0.6,
                                           num_return_sequences=args.n_solutions_sampled)
            output_new_tokens = entire_output[:, eval_batch["input_ids"].size(1):]
            decoded = tokenizer.batch_decode(output_new_tokens, skip_special_tokens=True)

            for j, ex in enumerate(examples[i:i + args.test_per_device_train_batch_size]):
                predictions = []
                predictions_accuracies = []
                for prediction in decoded[j * args.n_solutions_sampled:(j + 1) * args.n_solutions_sampled]:
                    prediction = prediction.strip()
                    predictions.append(prediction)

                    accuracy = evaluator.compute_metrics(prediction, ex)["accuracy"]

                    predictions_accuracies.append(accuracy)

                eval_result.append({
                    "model": model_name,
                    "input": ex["question"],
                    "gold": ex["answer"].split("####")[1].strip(),
                    "prediction_0": predictions[0],
                    "pass@k": any(predictions_accuracies),
                    "pass_rate": np.mean(predictions_accuracies),
                })
                accuracies.append(any(predictions_accuracies))
            eval_loop.set_description(f"Evaluating, accuracy: {np.mean(accuracies):.3f}")
            i += args.test_per_device_train_batch_size
        return eval_result

    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)

    train_dataset = load_dataset("json", data_files={"train": "code_reasoning_ablations_gsm8k_code.jsonl"})["train"].map(lambda ex: {"answer": ex["python"]})
    train_dataset = train_dataset.shuffle(seed=seed).select(range(args.n_train_samples))

    accelerator = Accelerator()

    tokenized_datasets = {
        "train": train_dataset.map(
            tokenize_function_train,
            batched=False,
            remove_columns=[name for name in train_dataset.column_names if
                            name not in ["input_ids", "labels", "attention_mask"]],
        ),
        "test": test_dataset.map(
            tokenize_function_eval,
            batched=False,
            remove_columns=[name for name in test_dataset.column_names if
                            name not in ["input_ids", "labels", "attention_mask"]],
        ),
    }

    train_dataloader = DataLoader(
        tokenized_datasets["train"],
        shuffle=True,
        collate_fn=DataCollatorForSeq2Seq(tokenizer=tokenizer, model=model, padding="longest"),
        batch_size=args.per_device_train_batch_size
    )

    test_dataloader = DataLoader(
        tokenized_datasets["test"],
        shuffle=False,
        collate_fn=DataCollatorForSeq2Seq(tokenizer=tokenizer, model=model, padding="longest"),
        batch_size=args.test_per_device_train_batch_size
    )

    optimizer = torch.optim.AdamW([p for n, p in model.named_parameters()], lr=args.learning_rate)
    num_training_steps_for_scheduler = len(train_dataloader) * args.num_train_epochs

    lr_scheduler = get_scheduler(
        name="linear",
        optimizer=optimizer,
        num_training_steps=num_training_steps_for_scheduler,
        num_warmup_steps=int(num_training_steps_for_scheduler * args.warmup_ratio),
    )

    # Prepare everything with `accelerator`.
    model, optimizer, train_dataloader, test_dataloader, lr_scheduler = accelerator.prepare(
        model, optimizer, train_dataloader, test_dataloader, lr_scheduler
    )

    acc_per_epoch = []
    results_per_epoch = []

    for epoch in range(args.num_train_epochs):
        model.train()
        total_loss = 0

        loop = tqdm(train_dataloader)
        for step, batch in enumerate(loop):
            with accelerator.accumulate(model):
                outputs = model(**batch, use_cache=False)
                loss = outputs.loss

                total_loss += loss.detach().float()
                accelerator.backward(loss)

                loop.set_description(f"Epoch {epoch}, loss={total_loss / (step + 1):.3f}")

                optimizer.step()
                optimizer.zero_grad()
                lr_scheduler.step()

        model.eval()

        eval_results = evaluate(model, tokenizer, test_dataset, test_dataloader)
        acc = pd.DataFrame(eval_results)["pass@k"].mean()
        print(f"accuracy epoch {epoch}: {acc}")

        acc_per_epoch.append(acc)
        results_per_epoch.append(eval_results)

    best_acc = max(acc_per_epoch)
    best_epoch = acc_per_epoch.index(best_acc)

    agg_results = [{
            "model": model_name,
            "seed": seed,
            "pass@k": max(acc_per_epoch)}]
    return results_per_epoch[best_epoch], agg_results


def main():
    args = get_args()

    random.seed(args.seed)

    hf_dataset = {"babi": ("Muennighoff/babi", ), "web_nlg": ("GEM/web_nlg", "en"), "gsm8k": ("gsm8k", "main")}[args.dataset]

    print("Loading dataset...")
    train_dataset = load_dataset(*hf_dataset)["train"]
    test_dataset = load_dataset(*hf_dataset)["test"]
    print("Dataset loaded.")

    if args.n_test_samples is not None:
        sampled_indices = random.sample(range(len(test_dataset)), args.n_test_samples)
        test_dataset = test_dataset.select(sampled_indices)

    if args.n_train_samples is not None:
        sampled_indices = random.sample(range(len(train_dataset)), args.n_train_samples)
        train_dataset = train_dataset.select(sampled_indices)

    evaluator = {"babi": BabiEval, "web_nlg": WebNLGEval, "gsm8k": GSM8KEval}[args.dataset]()

    checkpoints_of_evaluated_models = models_checkpoints if args.models is None else {model_name: checkpoint for
                                                                                      model_name, checkpoint in
                                                                                      models_checkpoints.items() if
                                                                                      model_name in args.models}

    results, agg_results = [], []
    for model_name, checkpoint in checkpoints_of_evaluated_models.items():
        print(f"*** Loading {model_name} ***")
        model = OLMoForCausalLM.from_pretrained(checkpoint)
        tokenizer = OLMoTokenizerFast.from_pretrained(checkpoint)

        tokenizer.padding_side = "left"
        tokenizer.pad_token = tokenizer.eos_token
        print("Model loaded.")

        for seed in range(args.number_seeds):
            if args.dataset in ["babi", "web_nlg"]:
                seed_results, seed_agg_results = eval_only(evaluator, model, model_name, tokenizer, train_dataset, test_dataset, seed, args)
            else:
                seed_results, seed_agg_results = train_and_eval_gsm8k(evaluator, model, model_name, tokenizer, test_dataset, seed, args)
            results += seed_results
            agg_results += seed_agg_results

    timestamp = pd.Timestamp.now().strftime("%Y%m%d_%H%M%S")

    dir_path = os.path.join("results", args.dataset)
    os.makedirs(dir_path, exist_ok=True)

    df = pd.DataFrame(results)
    filepath = os.path.join(dir_path, f"results_{timestamp}.csv")
    print(f"Saving results to {filepath}")
    df.to_csv(filepath, index=False)

    df_agg = pd.DataFrame(agg_results)
    filepath = os.path.join(dir_path, f"results_agg_{timestamp}.csv")
    print(f"Saving aggregated results to {filepath}")
    df_agg.to_csv(filepath, index=False)

    # final results
    metric_keys = [key for key in df_agg.keys() if key not in ["model", "seed", "few_shot"]]
    print(df.groupby(['model']).mean(numeric_only=True)[metric_keys])

if __name__ == '__main__':
    main()



"""A Dataflow script for creating DOLMA formatted pretraining data from
reddit comment dumps.

For usage see README.md.

Adapted from work done at PolyAI:
https://github.com/PolyAI-LDN/conversational-datasets/tree/master/reddit
"""

import logging
from functools import partial
import uuid
import apache_beam as beam
from apache_beam.options.pipeline_options import PipelineOptions
from utils.shared_utils import (
    read_content_from_source,
    build_base_parser,
    positive_int,
    normalize_id,
    normalize_string,
    load_filtered_subreddit_lists,
    write_to_gcs,
    isodatetime_from_epoch,
    DATA_ACQUISITION_DATE
)


def parse_args(argv=None):
    parser = build_base_parser()

    parser.add_argument(
        "--parent_depth",
        type=positive_int,
        default=12,
        help="How many parent content to consider.",
    )
    parser.add_argument(
        "--max_length",
        type=positive_int,
        default=40000,
        help="Maximum letters in content to include.",
    )
    parser.add_argument(
        "--min_length",
        type=positive_int,
        default=500,
        help="Minimum characters in content to include.",
    )
    return parser.parse_known_args(argv)


def normalize_comment(comment, max_length):

    def trim(text, max_length):
        if len(text) <= max_length:
            return text
        text = text[:max_length + 1]

        # Trim until the last two characters are the boundary between an
        # alphanumeric character, and a non-alphanumeric character.
        while len(text) > 1 and (text[-1].isalnum() == text[-2].isalnum()):
            text = text[:-1]
        return text[:-1]

    comment = {
        'id': comment['id'],
        'thread_id': normalize_id(comment['link_id']),
        'parent_id': normalize_id(comment['parent_id']),
        'body': trim(normalize_string(comment['body']), max_length),
        'body_is_trimmed': len(comment['body']) > max_length,
        'author': normalize_string(comment['author']),
        'subreddit': normalize_string(comment['subreddit']),
        'created_utc': comment['created_utc'],
        'score': comment['score'],
        'link_id': comment['link_id']
    }
    return comment


def create_examples(comment, min_length):
    def _should_skip(comment, min_length):
        if comment['body_is_trimmed']:
            return True
        if comment['body'] in {
            "[deleted]",
            "[removed]",
            "[UNICODE ENCODE ERROR]"}:
            return True
        if comment['subreddit'] in {
            "[deleted]",
            "[removed]",
            "[UNICODE ENCODE ERROR]"}:
            return True
        if len(comment['body']) < min_length:
            return True
        if not comment['score'] or comment['score'] < 2:
            return True
        return False

    if _should_skip(comment, min_length):
        return None

    example = {
        'subreddit': comment['subreddit'],
        'thread_id': comment['thread_id'],
        'created': isodatetime_from_epoch(comment['created_utc']),
        'added': DATA_ACQUISITION_DATE,
        'id': uuid.uuid4().hex,
        'conversational_format': comment['body']
    }
    yield example


def run(argv=None, comments=None):
    args, pipeline_args = parse_args(argv)
    banned_subreddits_file = args.banned_subreddits_file
    banned_subreddits = load_filtered_subreddit_lists(banned_subreddits_file)

    pipeline_options = PipelineOptions(pipeline_args, save_main_session=True,
                                       dataflow_service_options=['enable_prime'],
                                       experiments=['enable_batch_vmr',
                                                    'enable_vertical_memory_autoscaling',
                                                    'no_use_multiple_sdk_containers'])

    p = beam.Pipeline(options=pipeline_options)

    comments = read_content_from_source(comments, p, args)

    comments |= (
        "normalize content" >> beam.Map(
            partial(normalize_comment, max_length=args.max_length)))

    examples = comments | (
        "Create examples" >> beam.FlatMap(
            partial(create_examples,
                    min_length=args.min_length,
                    )))

    examples = examples | (
            "Filter none content" >> beam.Filter(
        lambda c: c is not None))

    write_to_gcs(examples, banned_subreddits, args)

    result = p.run()
    result.wait_until_finish()


if __name__ == "__main__":
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    run()



from typing import Callable, Dict, Generator, Tuple, Type, TypeVar

from .taggers import BaseTagger

T = TypeVar("T", bound=BaseTagger)


class TaggerRegistry:
    """A registry for taggers."""

    __taggers: Dict[str, Type[BaseTagger]] = {}

    @classmethod
    def taggers(cls) -> Generator[Tuple[str, Type[BaseTagger]], None, None]:
        """Yield all taggers."""
        yield from cls.__taggers.items()

    @classmethod
    def add(cls, name: str) -> Callable[[Type[T]], Type[T]]:
        """Add a tagger to the registry."""

        def _add(
            tagger_cls: Type[T],
            tagger_name: str = name,
            taggers_dict: Dict[str, Type[BaseTagger]] = cls.__taggers,
        ) -> Type[T]:
            """Add a tagger to the registry using tagger_name as the name."""
            if tagger_name in taggers_dict and taggers_dict[tagger_name] != tagger_cls:
                if tagger_cls.__module__ == "__main__":
                    return tagger_cls

                raise ValueError(f"Tagger {tagger_name} already exists")

            taggers_dict[tagger_name] = tagger_cls
            return tagger_cls

        return _add

    @classmethod
    def remove(cls, name: str) -> bool:
        """Remove a tagger from the registry."""
        if name in cls.__taggers:
            cls.__taggers.pop(name)
            return True
        return False

    @classmethod
    def has(cls, name: str) -> bool:
        """Check if a tagger exists in the registry."""
        return name in cls.__taggers

    @classmethod
    def get(cls, name: str) -> Type[BaseTagger]:
        """Get a tagger from the registry; raise ValueError if it doesn't exist."""
        if name not in cls.__taggers:
            raise ValueError(
                f"Unknown tagger {name}; available taggers: " + ", ".join([tn for tn, _ in cls.taggers()])
            )
        return cls.__taggers[name]



"""A Dataflow script for creating DOLMA formatted pretraining data from
reddit submission dumps.

For usage see README.md.

Adapted from work done at PolyAI:
https://github.com/PolyAI-LDN/conversational-datasets/tree/master/reddit
"""

import logging
from functools import partial
import apache_beam as beam
from apache_beam.options.pipeline_options import PipelineOptions
from utils.shared_utils import (
    build_base_parser,
    positive_int,
    normalize_id,
    normalize_string,
    load_filtered_subreddit_lists,
    read_content_from_source,
    write_to_gcs,
    DATA_ACQUISITION_DATE
)

def parse_args(argv=None):
    parser = build_base_parser()

    parser.add_argument(
        "--max_length",
        type=positive_int,
        default=10000,
        help="Maximum words in posts to include.",
    )
    parser.add_argument(
        "--min_length",
        type=positive_int,
        default=60,
        help="Minimum words in posts to include.",
    )
    parser.add_argument(
        "--min_score",
        default=2,
        type=positive_int,
        help="The minimum score of a post to include.",
    )
    return parser.parse_known_args(argv)


def normalize_post(post, max_length):

    body_key = "body" if "body" in post else "selftext"
    post = {
        "id": post['id'],
        "title": normalize_string(post['title']) if 'title' in post else None,
        "author": normalize_string(post['author']) if 'author' in post else None,
        "subreddit": normalize_string(post['subreddit']) if 'subreddit' in post else None,
        "subreddit_id": normalize_id(post['subreddit_id']) if 'subreddit_id' in post else None,
        "body": normalize_string(post[body_key]),
        "body_is_trimmed": len(post[body_key]) > max_length,
        "created_utc": post['created_utc'] if 'created_utc' in post else None,
        "score": post['score'] if 'score' in post else None,
        "over_18": post['over_18'] if 'over_18' in post else None,
        "num_comments": post['num_comments'] if 'num_comments' in post else None,
    }
    return post


def create_examples(post, min_length, min_score):

    def _should_skip(post, min_length, min_score):

        for key in post:
            if post[key] is None:
                return True

        if post['body_is_trimmed']:
            return True
        if post['body'] in {
                "[deleted]",
                "[removed]",
                "[UNICODE ENCODE ERROR]"} or post['author'] in {
                "[deleted]",
                "[removed]",
                "[UNICODE ENCODE ERROR]"}:
            return True
        if post['subreddit'] in {
            "[deleted]",
            "[removed]",
                "[UNICODE ENCODE ERROR]"}:
            return True
        if len(post['body']) < min_length or post['over_18'] or post['score'] < min_score:
            return True
        return False

    if _should_skip(post, min_length, min_score):
        return None

    example = {
        "id": post['id'],
        "title": post['title'],
        "author": post['author'],
        "subreddit": post['subreddit'],
        "body": post['body'],
        "score": post['score'],
        "over_18": post['over_18'],
        "num_comments": post['num_comments'],
        'added': DATA_ACQUISITION_DATE

    }
    if post['created_utc'] is None:
        example["created"] = None
    else:
        try:
            example["created"] = int(post['created_utc'])
        except ValueError:
            example["created"] = None

    example["created"] = post['created_utc']

    formatted_post_str = ''
    if example["title"] is not None:
        formatted_post_str += f"Title: {example['title']}\n\n\n"
    formatted_post_str += example['body']
    example['formatted_post'] = formatted_post_str

    yield example


def run(argv=None, posts=None):

    args, pipeline_args = parse_args(argv)
    banned_subreddits_file = args.banned_subreddits_file
    banned_subreddits = load_filtered_subreddit_lists(banned_subreddits_file)

    pipeline_options = PipelineOptions(pipeline_args, save_main_session=True)
    p = beam.Pipeline(options=pipeline_options)

    posts = read_content_from_source(posts, p, args)

    posts |= (
        "normalize posts" >> beam.Map(
            partial(normalize_post, max_length=args.max_length)))

    posts = posts | ("Filtering posts that are skipped" >>
                     beam.Filter(lambda post: post is not None))

    examples = posts | (
        "Create examples" >> beam.FlatMap(
            partial(create_examples,
                    min_length=args.min_length,
                    min_score=args.min_score,
                    )))

    write_to_gcs(examples, banned_subreddits, args)

    result = p.run()
    result.wait_until_finish()


if __name__ == "__main__":
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    run()



from __future__ import annotations

import json
import os
import re
from enum import Enum
from functools import cached_property
from itertools import chain
from os import PathLike
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Generator, List, Optional, Tuple, Union

import msgspec
import smart_open
from omegaconf import DictConfig
from omegaconf.omegaconf import OmegaConf as om
from tokenizers import Tokenizer as BaseTokenizer

from ..core.errors import DolmaConfigError
from ..core.loggers import get_logger
from .data_types import InputSpec, TokenizerOutput

PathOrStr = Union[str, PathLike]

logger = get_logger(__name__)


__all__ = ["Tokenizer"]


class StrEnum(str, Enum):
    """
    This is equivalent to Python's :class:`enum.StrEnum` since version 3.11.
    We include this here for compatibility with older version of Python.
    """

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return f"'{str(self)}'"


class TruncationDirection(StrEnum):
    right = "right"
    left = "left"


class Tokenizer:
    """
    A :class:`Tokenizer` is a light-weight wrapper around a HuggingFace :class:`tokenizers.Tokenizer`.

    :param base_tokenizer: The :class:`tokenizers.Tokenizer` to use.
    :param eos_token_id: The token ID corresponding to the "end-of-sentence" token.
    :param truncate_to: Truncate when tokenizing to this number of token IDs.
    :param truncate_direction: The direction to truncate in. "right" means truncate the tokens
        on the right. "left" means truncate the tokens on the left. If ``truncate_to`` is null,
        this setting has no effect.
    """

    def __init__(
        self,
        base_tokenizer: BaseTokenizer,
        bos_token_id: Optional[int] = None,
        eos_token_id: Optional[int] = None,
        pad_token_id: Optional[int] = None,
        truncate_to: Optional[int] = None,
        truncate_direction: Union[str, TruncationDirection] = TruncationDirection.right,
        segment_before_tokenization: bool = False,
    ):
        self.base_tokenizer = base_tokenizer
        self.base_tokenizer.no_truncation()
        self.bos_token_id = bos_token_id
        self.eos_token_id = eos_token_id
        self.pad_token_id = pad_token_id if pad_token_id is not None else eos_token_id

        if self.pad_token_id:
            logger.warning("No pad token ID provided; using 0.")
            self.pad_token_id = 0

        self.truncate_to = truncate_to
        self.truncate_direction = TruncationDirection(truncate_direction)
        self.segment_before_tokenization = segment_before_tokenization

        self.config = self.get_base_tokenizer_config()

    @cached_property
    def tokenizer_has_prefix(self) -> bool:
        """Returns true if the tokenizer has a prefix space. Used to determine if we need to add a space before
        tokenizer when segment_before_tokenization is True."""

        # if the tokenizer adds a prefix space, we much return True
        pretokenizer_config: dict = self.config.get("pre_tokenizer") or {}
        if pretokenizer_config.get("type") == "Sequence":
            # it's a sequence of pretokenizers, so we gotta check each one
            for pretok in pretokenizer_config.get("pretokenizers", []):
                if pretok.get("add_prefix_space", False):
                    return True
        elif pretokenizer_config.get("add_prefix_space", False):
            # this covers the case where the pretokenizer is a single pretokenizer
            return True

        # check if the normalizer or one of the components of the normalizer appends a prefix
        normalizer_config: dict = self.config.get("normalizer") or {}
        if normalizer_config.get("type") == "Sequence":
            # it's a sequence of normalizers, so we gotta check each one
            for norm in normalizer_config.get("normalizers", []):
                if norm.get("type", None) == "Prepend":
                    return True
        elif normalizer_config.get("type", None) == "Prepend":
            # this covers the case where the normalizer is a single normalizer
            return True

        # all checks above failed, so we return False
        return False

    def get_base_tokenizer_config(self) -> dict:
        # Rust HuggingFace tokenizers don't have a way to get the full configuration through Python bindings,
        # so we hack around it by saving the tokenizer to a temporary file and reading the config.
        with NamedTemporaryFile() as f:
            self.base_tokenizer.save(f.name)
            f.flush()
            f.seek(0)
            config = json.load(f)
        return config

    @property
    def vocab_size(self) -> int:
        return self.base_tokenizer.get_vocab_size()

    @classmethod
    def from_train_config(cls, config: DictConfig) -> "Tokenizer":
        tokenizer_identifier = config.tokenizer.identifier
        if Path(tokenizer_identifier).is_file():
            tokenizer = cls.from_file(
                tokenizer_identifier,
                eos_token_id=config.model.eos_token_id,
                pad_token_id=config.model.pad_token_id,
            )
        else:
            tokenizer = cls.from_pretrained(
                tokenizer_identifier,
                eos_token_id=config.model.eos_token_id,
                pad_token_id=config.model.pad_token_id,
            )
        if config.model.vocab_size != tokenizer.vocab_size:
            raise DolmaConfigError("vocab size mismatch between config and tokenizer")
        return tokenizer

    @classmethod
    def from_pretrained(cls, identifier: str, **kwargs) -> "Tokenizer":
        """
        Initialize a tokenizer from a pretrained tokenizer on the HuggingFace Hub.

        :param identifier: The identifier of a model on the Hub that contains a
            ``tokenizer.json`` file.
        :param kwargs: Other key word arguments passed to :class:`Tokenizer`.
        """
        base_tokenizer = BaseTokenizer.from_pretrained(identifier)
        return cls(base_tokenizer=base_tokenizer, **kwargs)

    @classmethod
    def from_file(cls, filename: PathOrStr, **kwargs) -> "Tokenizer":
        """
        Initialize a tokenizer from a file.

        You can create those files with ``BaseTokenizer.save()``.

        :param filename: The name of a file containing a tokenizer specification.
        :param kwargs: Other key word arguments passed to :class:`Tokenizer`.
        """
        base_tokenizer = BaseTokenizer.from_file(filename)
        return cls(base_tokenizer=base_tokenizer, **kwargs)

    @classmethod
    def from_checkpoint(cls, checkpoint_dir: PathOrStr) -> "Tokenizer":
        """
        Load a tokenizer from a checkpoint.
        """
        from cached_path import cached_path

        # Load configs.
        config_path = cached_path(os.path.join(checkpoint_dir, "config.yaml"))
        tokenizer_config = om.load(config_path).tokenizer
        model_config = om.load(config_path).model

        # Initialize tokenizer and validate vocab size.
        tokenizer = cls.from_pretrained(
            tokenizer_config.identifier,
            eos_token_id=model_config.eos_token_id,
            pad_token_id=model_config.pad_token_id,
        )
        if model_config.vocab_size != tokenizer.vocab_size:
            raise DolmaConfigError("vocab size mismatch between config and tokenizer")
        return tokenizer

    def add_special_tokens(self, input_ids: List[int]) -> List[int]:
        """
        Add special tokens in-place (if not already present) to the given token IDs.
        """
        if not input_ids:
            return input_ids

        if self.bos_token_id is not None and input_ids[0] != self.bos_token_id:
            input_ids.insert(0, self.bos_token_id)

        if self.eos_token_id is not None and input_ids[-1] != self.eos_token_id:
            input_ids.append(self.eos_token_id)

        return input_ids

    def num_special_tokens_to_add(self) -> int:
        return (1 if self.eos_token_id is not None else 0) + (1 if self.bos_token_id is not None else 0)

    def _truncate(
        self, input_ids: List[int], truncate_to: Optional[int], direction: TruncationDirection
    ) -> list[int]:
        if truncate_to is None or len(input_ids) <= truncate_to:
            return input_ids
        elif direction == TruncationDirection.left:
            return input_ids[len(input_ids) - truncate_to :]
        else:
            return input_ids[: -(len(input_ids) - truncate_to)]

    def encode(self, input: str, add_special_tokens: bool = True) -> List[int]:
        """
        Encode a string into token IDs.
        """
        return self.encode_batch([input], add_special_tokens=add_special_tokens)[0]

    def split_into_paragraphs(self, inputs: List[str]) -> Tuple[List[str], List[Tuple[int, int]]]:
        slices = []
        batch = []
        curr = 0
        for input in inputs:
            paragraphs = [
                # if a tokenizer adds a prefix in front of sequences, then the tokenization of the first
                # symbol in each paragraph will be different depending on whether paragraphs are split
                # before tokenization or not. To counter this, we add a space in front of each paragraph
                # except the first one. We will remove the space from the tokenized symbols later.
                (" " if self.tokenizer_has_prefix and i > 0 else "") + match.group()
                # this regular expression keeps newlines at the beginning of paragraphs unless
                # the paragraph is the first one in the document
                for i, match in enumerate(re.finditer(r"(^\n*|\n+)[^\n]*", input))
            ]
            slices.append((curr, curr + len(paragraphs)))
            batch.extend(paragraphs)
            curr += len(paragraphs)
        return batch, slices

    def merge_paragraphs(self, encoded: List[List[int]], slices: List[Tuple[int, int]]) -> List[List[int]]:
        merged = []
        for start, end in slices:
            encoded_slice_iter = (
                # the slicing operation is required if we have added a space in front of each paragraph
                # during the `split_into_paragraphs` method.
                encoded[pos][1:] if (self.tokenizer_has_prefix and pos > 0) else encoded[pos]
                for pos in range(start, end)
            )
            merged.append(list(chain.from_iterable(encoded_slice_iter)))
        return merged

    def encode_batch(self, inputs: List[str], add_special_tokens: bool = True) -> List[List[int]]:
        """
        Encode a batch of strings into token IDs.
        """
        truncate_to = self.truncate_to
        if truncate_to is not None and add_special_tokens:
            truncate_to -= self.num_special_tokens_to_add()

        if self.segment_before_tokenization:
            sliced_inputs, slice_locs = self.split_into_paragraphs(inputs)
            sliced_batch_encoding = [
                e.ids for e in self.base_tokenizer.encode_batch(sliced_inputs, add_special_tokens=False)
            ]
            batch_encoding = self.merge_paragraphs(sliced_batch_encoding, slice_locs)
        else:
            batch_encoding = [e.ids for e in self.base_tokenizer.encode_batch(inputs, add_special_tokens=False)]

        all_input_ids = []
        for encoding in batch_encoding:
            input_ids = self._truncate(encoding, truncate_to, self.truncate_direction)
            if add_special_tokens:
                input_ids = self.add_special_tokens(input_ids)
            all_input_ids.append(input_ids)

        return all_input_ids

    def decode(self, token_ids: List[int], skip_special_tokens: bool = True) -> str:
        """
        Decode a list of token IDs to a string.
        """
        return self.base_tokenizer.decode(token_ids, skip_special_tokens=skip_special_tokens)


def tokenize_file(tokenizer: Tokenizer, path: str) -> Generator[TokenizerOutput, None, None]:
    """Tokenize a file of documents using the provided tokenizer; file is expected to be a gzipped JSON lines
    file, each containing a field named `text`.
    """
    decoder = msgspec.json.Decoder(InputSpec)
    with smart_open.open(path, mode="rt") as input_stream:
        for i, line in enumerate(input_stream, start=1):
            try:
                row = decoder.decode(line)
                if text := row.text.strip():
                    # skip empty docs
                    tokens = tokenizer.encode(text, add_special_tokens=True)
                    yield TokenizerOutput.from_tokens(id=row.id, src=path, loc=i, tokens=tokens)
                i += 1
            except Exception as ex:
                logger.error("Error processing %s:%d", path, i, exc_info=ex)



import copy
import os
import tempfile
from contextlib import ExitStack, contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Generator, Optional

from dolma.cli import field


@dataclass
class WorkDirConfig:
    input: Optional[str] = field(default=None, help="Path to the input directory.")
    output: Optional[str] = field(default=None, help="Path to the output directory.")


@contextmanager
def get_path_to_temp_file(prefix="dolma-", suffix=None) -> Generator[Path, None, None]:
    with tempfile.NamedTemporaryFile(prefix=prefix, suffix=suffix, delete=True) as f:
        path = Path(f.name)
    yield path

    if path.exists():
        os.remove(path)


@contextmanager
def make_workdirs(config: WorkDirConfig) -> Generator[WorkDirConfig, None, None]:
    """Create temporary work directories and update the config with their paths."""

    # make a copy of the configuration
    config = copy.deepcopy(config)

    with ExitStack() as stack:
        if config.input is None:
            config.input = stack.enter_context(tempfile.TemporaryDirectory(prefix="dolma-input-"))
        if config.output is None:
            config.output = stack.enter_context(tempfile.TemporaryDirectory(prefix="dolma-output-"))

        yield config



import math
from abc import abstractmethod, abstractproperty
from typing import Dict, List, NamedTuple, Optional, Tuple, Union

import numpy as np
import numpy.typing as npt

# # # OLD IMPORT # # #
# from sortedcontainers import SortedDict


class SummaryTuple(NamedTuple):
    counts: List[int]
    bins: List[float]


def sort_and_merge_bins(
    bins: npt.NDArray[np.float64], counts: npt.NDArray[np.int64], mask: Optional[npt.NDArray[np.bool_]] = None
) -> Tuple[npt.NDArray[np.float64], npt.NDArray[np.int64]]:
    """Sort bins and counts; merge counts for duplicate bins"""

    masked_bins = bins[mask] if mask is not None else bins
    masked_counts = counts[mask] if mask is not None else counts

    uniq_bins, uniq_indices, uniq_counts = np.unique(masked_bins, return_counts=True, return_index=True)
    uniq_counts *= masked_counts[uniq_indices]

    return uniq_bins, uniq_counts


def merge_bins(
    bin_a: npt.NDArray[np.float64],
    count_a: npt.NDArray[np.int64],
    bin_b: npt.NDArray[np.float64],
    count_b: npt.NDArray[np.int64],
):
    """Merge two sets of bins and counts into one set of bins and counts;
    assumes that bin_a and bin_b are sorted

    Args:
        bin_a (npt.NDArray[np.float64]): A sorted array of bins
        count_a (npt.NDArray[np.int64]): A corresponding array of counts
        bin_b (npt.NDArray[np.float64]): A sorted array of bins
        count_b (npt.NDArray[np.int64]): A corresponding array of counts
    """
    if bin_a.size < bin_b.size:
        # bin_a is always the larger one
        bin_a, count_a, bin_b, count_b = bin_b, count_b, bin_a, count_a

    # we first find where the bins in bin_b would be inserted into bin_a
    # b_locs = np.minimum(np.searchsorted(bin_a, bin_b, side="left"), bin_a.size - 1)
    b_locs = np.searchsorted(bin_a, bin_b, side="left")

    # make a masked version of b_locs that only contains indices that are in bounds
    b_bounded_mask = b_locs < bin_a.size
    b_bounded_locs = b_locs * b_bounded_mask.astype(b_locs.dtype)

    # we make a few useful masks and arrays for later operations
    # we need to keep track of which bins in bin_b are new and which are duplicates of bins in bin_a
    # for the former, we will insert them into bin_a at the appropriate locations; for the latter,
    # we will add their counts to the counts of the corresponding bins in bin_a

    # new mask consists of either the bins in bin_b that are in bounds (i.e., would be between values of
    # bin_a and not after) and are not equal to the corresponding bins in bin_a, or the bins in bin_b that
    # will be inserted after the last bin in bin_a.
    b_new_mask = (bin_a[b_bounded_locs] != bin_b) | ~b_bounded_mask
    b_new_vals = bin_b[b_new_mask]

    # now we need to find the locations where the new bins will be inserted into new array bin_c which
    # is the size of bin_a + the number of new bins in bin_b
    b_new_locs = np.arange(b_new_vals.size) + b_locs[b_new_mask]

    # this is were we will store the new bins and counts
    bin_c = np.empty(bin_a.size + b_new_vals.size, dtype=bin_a.dtype)

    # we first fill bins from bin_a into bin_c; a_indices is a mask of the indices in bin_c that
    # should be filled with values from bin_a, so we remove values from bin_c.
    a_indices = np.ones(bin_c.size, dtype=bool)
    a_indices[b_new_locs] = False
    bin_c[a_indices] = bin_a

    # finally, we add values from bin_b into bin_c
    bin_c[b_new_locs] = b_new_vals

    # now onto the counts; we start by creating new container, and populate counts from count_a and count_b
    # where bin_a values are different from bin_b values.
    count_c = np.empty_like(bin_c, dtype=count_a.dtype)
    count_c[b_new_locs] = count_b[b_new_mask]
    count_c[a_indices] = count_a

    # finally, for the remaining counts, we group them by bin value and add them to counts from count_a
    # we must group because `array[locs] += values` does not work if there are duplicate indices in locs.
    b_uniq_locs, b_repeats, b_rep_cnt = np.unique(b_locs[~b_new_mask], return_counts=True, return_index=True)
    count_c[b_uniq_locs] += count_b[~b_new_mask][b_repeats] * b_rep_cnt

    return bin_c, count_c


class BaseBucketApi:
    @abstractproperty
    def full(self) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def add(self, value: Union[int, float], count: int = 1):
        raise NotImplementedError()

    def add_many(self, values: List[Union[int, float]], counts: List[int]):
        for value, count in zip(values, counts):
            self.add(value, count)

    @abstractmethod
    def summarize(self, n: int, density: bool = False) -> SummaryTuple:
        raise NotImplementedError()


class InferBucketsValTracker(BaseBucketApi):
    """Keep track of running values by using two bucketed buffers"""

    _bins: npt.NDArray[np.float64]
    _counts: npt.NDArray[np.int64]
    _buffer_bins: npt.NDArray[np.float64]
    _buffer_counts: npt.NDArray[np.int64]
    _buffer_idx: int
    n: int
    _n: int

    def __init__(self, n: int, b: Optional[int] = None):
        self.n = self._n = n
        self.b = b or int(np.sqrt(n))

        self._bins = np.empty(0, dtype=np.float64)
        self._counts = np.empty_like(self._bins, dtype=np.int64)

        # hold temporary values in a buffer
        self._new_buffer()

    def _new_buffer(self):
        """Create a new buffer and reset the buffer index"""
        self._buffer_bins = np.zeros(self.b, dtype=np.float64)
        self._buffer_counts = np.zeros_like(self._buffer_bins, dtype=np.int64)
        self._buffer_idx = 0

    def _add_buffer_to_bins(self):
        """Bin the values in the buffer and merge them with the existing bins and counts."""
        locs = np.minimum(np.searchsorted(self._bins, self._buffer_bins, side="left"), self._bins.size - 1)
        trim_og_locs, trim_og_counts = locs[: self._buffer_idx], self._buffer_counts[: self._buffer_idx]
        trim_locs, repeats_locs, repeats_cnt = np.unique(trim_og_locs, return_counts=True, return_index=True)
        trim_counts = trim_og_counts[repeats_locs] * repeats_cnt
        self._counts[trim_locs] += trim_counts

        self._new_buffer()

    def _sort_buffer(self):
        mask = np.arange(0, self._buffer_bins.size) < self._buffer_idx
        bins, counts = sort_and_merge_bins(bins=self._buffer_bins, counts=self._buffer_counts, mask=mask)

        self._buffer_bins = bins
        self._buffer_counts = counts
        self._buffer_idx = bins.size

    def _concat_buffer(self):
        """Concatenate the buffer with the existing bins and counts."""

        # make sure the buffer is sorted before merging
        self._sort_buffer()

        if self._bins.size == 0:
            # shortcut: if there are no bins, just copy the buffer
            self._bins = self._buffer_bins
            self._counts = self._buffer_counts
            self._new_buffer()
            return

        # actually do the merge here!
        self._bins, self._counts = merge_bins(
            bin_a=self._bins, count_a=self._counts, bin_b=self._buffer_bins, count_b=self._buffer_counts
        )
        self._new_buffer()

    def _add_not_full(self, value: float, count: int = 1):
        """Add a value to the tracker when the tracker is not full; in this case, the value is
        added to the buffer and eventually merged with existing bins and counts."""

        self._n -= 1
        if self._n < 0:
            return self._add_full(value, count)

        self._buffer_bins[self._buffer_idx] = value
        self._buffer_counts[self._buffer_idx] = count
        self._buffer_idx += 1

        if self._buffer_idx == self._buffer_bins.size:
            self._concat_buffer()

    def _add_full(self, value: float, count: int = 1):
        """Add a value to the tracker when the tracker is full; in this case, the value is added by
        bisecting the tracker and adding the value to the appropriate bucket."""
        self._buffer_bins[self._buffer_idx] = value
        self._buffer_counts[self._buffer_idx] = count
        self._buffer_idx += 1

        if self._buffer_idx == self._buffer_bins.size:
            self._add_buffer_to_bins()

    def __len__(self) -> int:
        return self._counts.size

    @property
    def full(self) -> bool:
        return self._n <= 0

    def add(self, value: Union[int, float], count: int = 1):
        if self._n >= 0:
            self._add_not_full(value=value, count=count)
        else:
            self._add_full(value=value, count=count)

    def summarize(self, n: int, density: bool = False) -> SummaryTuple:
        """Return up to n buckets with counts of merged values"""

        # finalize operations
        self._concat_buffer() if self._n >= 0 else self._add_buffer_to_bins()

        if len(self) <= n:
            # if there are fewer than n buckets, return the buckets as is
            return SummaryTuple(counts=self._counts.tolist(), bins=self._bins.tolist())

        # make weighted histogram using counts
        new_counts, new_values = np.histogram(a=self._bins, bins=n, weights=self._counts, density=density)

        # return lists instead of numpy arrays
        return SummaryTuple(counts=new_counts.tolist(), bins=new_values.tolist())


class FixedBucketsValTracker(BaseBucketApi):
    def __init__(self, n: int = 2):
        # we use n to determine the precision of the bins; for convenience we store it as a power of 10.
        # 10**n will be the maximum number of bins for each power of 2.
        # Too large numbers will cause numeric problems and can cause a lot of memory use.
        assert n >= 0
        assert n <= 100
        self.n = 10**n
        self._bins: Dict[Tuple[int, int], int] = {}

    def add(self, value: Union[int, float], count: int = 1):
        m, e = math.frexp(value)
        k = math.floor(m * self.n), e

        if k not in self._bins:
            self._bins[k] = 0
        self._bins[k] += count

    def __len__(self) -> int:
        return len(self._bins)

    @property
    def full(self) -> bool:
        return False

    def get_bin_upper_bound(self, val: float) -> float:
        """Return the upper bound of the bin containing val"""
        m, e = math.frexp(val)
        k = math.floor(m * self.n) + 1  # Add one to obtain the next bin
        return k / self.n * 2**e

    def summarize(self, n: int, density: bool = False) -> SummaryTuple:
        bins, counts = zip(*sorted((m / self.n * 2**e, c) for (m, e), c in self._bins.items()))

        if len(self) <= n:
            # if there are fewer than n buckets, return the buckets as is
            # To be consistent we also add the limit of the last bin, so the bins denote bin edges
            upper_bin = self.get_bin_upper_bound(max(float(b) for b in bins))
            return SummaryTuple(counts=[int(c) for c in counts], bins=[float(b) for b in bins] + [upper_bin])

        # computing the weighted histograms
        new_counts, new_values = np.histogram(a=bins, bins=n, weights=counts, density=density)

        # return lists instead of numpy arrays
        return SummaryTuple(counts=new_counts.tolist(), bins=new_values.tolist())



"""
Convert a dataset that is in the format used by the WIMBD project to the format used in Dolma.

Author: Luca Soldaini (@soldni)
"""

import argparse
import datetime
import hashlib
import json
import multiprocessing
from contextlib import ExitStack
from queue import Queue
from tempfile import TemporaryDirectory
from typing import Any, Dict, List, Tuple, Union

import msgspec
import smart_open

from dolma.core.parallel import BaseParallelProcessor
from dolma.core.paths import join_path
from dolma.core.runtime import _make_paths_from_prefix


class WimbdInputSpec(msgspec.Struct):
    text: str
    metadata: Dict[str, Any] = msgspec.field(default_factory=dict)


class DolmaInputSpec(msgspec.Struct):
    id: str
    text: str
    source: str
    version: str
    added: str
    created: str
    metadata: Dict[str, Any] = msgspec.field(default_factory=dict)


def convert_timestamp(d: datetime.datetime) -> str:
    return d.strftime("%Y-%m-%dT%H:%M:%S.%f")[:23] + "Z"


class WimbdToDolmaProcessor(BaseParallelProcessor):
    @classmethod
    def increment_progressbar(  # type: ignore
        cls,
        queue: "Queue[Union[Tuple[int, ...], None]]",
        /,
        files: int = 0,
        documents: int = 0,
    ) -> Dict[str, int]:
        return super().increment_progressbar(
            queue,
            files=files,
            documents=documents,
        )

    @classmethod
    def process_single(
        cls, source_path: str, destination_path: str, queue: "Queue[Union[Tuple[int, ...], None]]", **kwargs: Any
    ):
        decoder = msgspec.json.Decoder(WimbdInputSpec)
        encoder = msgspec.json.Encoder()

        source_name = kwargs.pop("source_name", None)
        source_version = kwargs.pop("source_version", "v0")
        assert isinstance(source_name, str), "source_name should be a string"

        documents_count = 0
        update_interval = 1

        current_time = convert_timestamp(datetime.datetime.now())

        with ExitStack() as stack:
            source = stack.enter_context(smart_open.open(source_path, "rt"))
            destination = stack.enter_context(smart_open.open(destination_path, "wb"))

            for line in source:
                data = decoder.decode(line)
                id_ = hashlib.md5(data.text.encode()).hexdigest()
                converted_doucument = DolmaInputSpec(
                    id=id_,
                    text=data.text,
                    source=source_name,
                    version=source_version,
                    added=current_time,
                    created=current_time,
                    metadata=data.metadata,
                )
                destination.write(encoder.encode(converted_doucument) + b"\n")  # type: ignore
                documents_count += 1

                if documents_count % update_interval == 0:
                    # update the progress bar every 1000 documents to prevent
                    # buffering
                    cls.increment_progressbar(queue, documents=documents_count)

                    if queue.qsize() >= multiprocessing.cpu_count():
                        # double the update interval if the queue is full
                        update_interval *= 2

        cls.increment_progressbar(queue, files=1, documents=documents_count)


def main(
    source: Union[List[str], str],
    destination: str,
    source_name: str,
    source_version: str = "v0",
    num_workers: int = 1,
    debug: bool = False,
) -> None:
    # make source always a list
    source = [source] if isinstance(source, str) else source
    with TemporaryDirectory() as tempdir:
        if len(source) > 1:
            dest_prefixes = _make_paths_from_prefix(source, join_path(None, destination))
            meta_prefixes = _make_paths_from_prefix(source, join_path(None, tempdir))
        else:
            dest_prefixes = [destination]
            meta_prefixes = [tempdir]

        processor = WimbdToDolmaProcessor(
            source_prefix=source,
            destination_prefix=dest_prefixes,
            metadata_prefix=meta_prefixes,
            num_processes=num_workers,
            debug=debug,
        )
        processor(source_name=source_name, source_version=source_version)


def parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument("-s", "--source", type=str, required=True, help="glob pattern for source files", nargs="+")
    ap.add_argument("-d", "--destination", type=str, required=True, help="destination prefix")
    ap.add_argument("-w", "--num-workers", type=int, default=1, help="number of workers")
    ap.add_argument("-n", "--source-name", required=True, help="source name")
    ap.add_argument("-v", "--source-version", default="v0", help="version")
    ap.add_argument("--debug", action="store_true", help="debug mode")
    opts = ap.parse_args()

    print(json.dumps(vars(opts), indent=2, sort_keys=True))
    return opts


if __name__ == "__main__":
    opts = parse_args()
    main(**vars(opts))



from .code_taggers import (
    CodeCopyrightTagger,
    CodeRedPajamaTaggers,
    CodeSecretsTagger,
    CodeStarCoderTaggers,
    CodeStarCoderTaggers2,
)

__all__ = [
    "CodeSecretsTagger",
    "CodeCopyrightTagger",
    "CodeRedPajamaTaggers",
    "CodeStarCoderTaggers",
    "CodeStarCoderTaggers2",
]



import math
import multiprocessing
import re
import shutil
from contextlib import ExitStack
from tempfile import TemporaryDirectory
from typing import Dict, List, Optional

import msgspec
import smart_open
import tqdm
from msgspec.json import Decoder
from rich.console import Console
from rich.table import Table

from .binning import BaseBucketApi, FixedBucketsValTracker, InferBucketsValTracker
from .data_types import OutputSpec
from .errors import DolmaError
from .parallel import BaseParallelProcessor, QueueType
from .paths import glob_path, mkdir_p

NUM_BINS = 100_000
BUFF_SIZE = 1_000


def _make_tracker(type_: str = "fixed", **kwargs: int) -> BaseBucketApi:
    """Make a tracker of given type. Choose between `infer` or `fixed`"""
    if type_ == "infer":
        return InferBucketsValTracker(**{"n": NUM_BINS, "b": BUFF_SIZE, **kwargs})
    elif type_ == "fixed":
        return FixedBucketsValTracker(**{"n": int(math.log10(NUM_BINS)), **kwargs})
    else:
        raise ValueError(f"Unknown tracker type {type_}")


class SummarySpec(msgspec.Struct):
    name: str
    counts: List[int]
    bins: List[float]

    @classmethod
    def from_tracker(cls, name: str, tracker: "BaseBucketApi", n: int) -> "SummarySpec":
        counts, bins = tracker.summarize(n=n)
        return SummarySpec(name=name, counts=counts, bins=bins)

    def to_tracker(self) -> "BaseBucketApi":
        tracker = _make_tracker()
        tracker.add_many(values=self.bins, counts=self.counts)
        return tracker


class AnalyzerProcessor(BaseParallelProcessor):
    @classmethod
    def increment_progressbar(  # type: ignore
        cls,
        queue: QueueType,  # queue must be the first argument, and it should be a positional-only argument
        /,
        files: int = 0,
        documents: int = 0,
    ) -> Dict[str, int]:
        """We override this method to specify which units we want to keep track of in a progress bar.
        Specifically, we keep track of files and documents in this example. Their default value must be zero."""

        # we call the super method to increment the progress bar
        return super().increment_progressbar(queue, files=files, documents=documents)

    @classmethod
    def process_single(
        cls,
        source_path: str,
        destination_path: str,
        queue: QueueType,
        **kwargs,
    ):
        # instantiate a decoder for faster decoding
        decoder = Decoder(OutputSpec)

        # number of bins to use
        num_bins = kwargs.get("num_bins", 1000)

        # regex to filter attribute names
        name_regex = re.compile(r) if (r := kwargs.get("name_regex", None)) else None

        # keep track of the length and score of each attribute
        trackers: Dict[str, BaseBucketApi] = {}

        # interval at which to update the progress bar; will double if queue is too full
        update_interval = 1

        # running document count; gets reset every time we update the progress bar
        docs_cnt = 0

        with smart_open.open(source_path) as f:
            for ln in f:
                try:
                    row = decoder.decode(ln)
                except Exception as e:
                    raise DolmaError(
                        f"Failed to decode line {ln} in {source_path}; "
                        f"are you sure {source_path} is an attributes file?"
                    ) from e

                # update the length and score trackers for each attribute
                for attr_name, attr_values in row.attributes.items():
                    # if a regex is provided, skip attributes that don't match it
                    if name_regex and not name_regex.search(attr_name):
                        continue

                    # empty attributes count as zero
                    attr_values = attr_values or [(0, 0, 0.0)]
                    for start, end, score in attr_values:
                        if "__label__" in attr_name:
                            # annoying fix for fasttext: fasttext sometimes emits probabilities that are slightly
                            # above 1.0, which causes issues with histograms. Therefore, we shift values that are
                            # greater than 1.0 down to 1.0
                            #
                            # fasttext labels are of the form __label__<label>, so we can just check if the
                            # attribute name contains __label__
                            score = min(score, 1.0)

                        trackers.setdefault(f"{attr_name}/score", _make_tracker()).add(score)
                        trackers.setdefault(f"{attr_name}/length", _make_tracker()).add(end - start)

                # increment the number of documents processed so far
                docs_cnt += 1

                if docs_cnt % update_interval == 0:
                    # update the progress bar every 1000 documents to prevent
                    # buffering
                    cls.increment_progressbar(queue, documents=docs_cnt)
                    docs_cnt = 0

                    if queue.qsize() >= multiprocessing.cpu_count():
                        # double the update interval if the queue is full
                        update_interval *= 2

        with smart_open.open(destination_path, "w") as f:
            for attr_name, tracker in trackers.items():
                summary = SummarySpec.from_tracker(name=attr_name, tracker=tracker, n=num_bins)
                f.write(msgspec.json.encode(summary).decode("utf-8") + "\n")

        # update the progress bar one last time
        cls.increment_progressbar(queue, files=1, documents=docs_cnt)


def aggregate_summaries(summaries_path: str, num_bins: int = 1000) -> List[SummarySpec]:
    # keep track of the length and score of each attribute
    trackers: Dict[str, BaseBucketApi] = {}

    # instantiate a decoder for faster decoding
    decoder = Decoder(SummarySpec)

    # iterator with nice progress bar
    it = tqdm.tqdm(list(glob_path(summaries_path)), desc="Aggregating summaries", unit=" files", unit_scale=True)

    # load partial summaries and aggregate it
    for path in it:
        with smart_open.open(path, "rt") as f:
            for ln in f:
                summary = decoder.decode(ln)
                trackers.setdefault(summary.name, _make_tracker()).add_many(summary.bins, summary.counts)

    # convert trackers to summaries
    summaries = [
        SummarySpec.from_tracker(name=attr_name, tracker=attr_tracker, n=num_bins)
        for attr_name, attr_tracker in trackers.items()
    ]
    return summaries


def visualize_summaries(summaries: List[SummarySpec], digits: int = 4, num_viz_bins: int = 10):
    console = Console()
    console.print()

    def round_all(values: List[float], opt_sci: bool = False) -> List[str]:
        """Logic to round values depending on their range"""

        if values == [0, 1]:
            # just 0 and 1; no need to round or add decimal points
            return ["0", "1"]
        elif all(-1 <= val <= 1 for val in values):
            # all values are in the range [-1, 1]; let's attempt rounding with {digits} decimal points
            # unless some values are identical after rounding.
            attempt_rounding = [round(val, digits) for val in values]

            if len(set(attempt_rounding)) != len(values) and opt_sci:
                # oops, some values collide after rounding; let's use scientific notation instead
                # with  one decimal point (note that we do this only if `opt_sci` is True)
                return [f"{val:.1e}" for val in values]
            else:
                # phew, all good; let's use {digits} decimal points for all values
                return [f"{round(val, digits):.{digits}f}" for val in values]
        else:
            # all values are outside the range [-1, 1]; let's round them to the nearest integer
            return [f"{int(round(val, 0)):d}" for val in values]

    for summary in summaries:
        # we use fewer bins for visualization
        summary = SummarySpec(
            name=summary.name,
            counts=(re_binned := summary.to_tracker().summarize(n=num_viz_bins)).counts,
            bins=re_binned.bins,
        )

        # build the table here
        table = Table(title=summary.name, style="bold", min_width=len(summary.name))
        table.add_column("value", justify="left", style="cyan")
        table.add_column("dist", justify="left", style="magenta")
        table.add_column("count", justify="left", style="green")

        rounded_bins = round_all(summary.bins)
        ranges = (
            [f"[{lo}, {hi})" for lo, hi in zip(rounded_bins, rounded_bins[1:])]
            if len(summary.bins) > len(summary.counts)
            else rounded_bins
        )

        counts_sum = sum(summary.counts)
        counts_normed = round_all([(count / counts_sum) for count in summary.counts], opt_sci=False)

        for value, dist, count in zip(ranges, counts_normed, summary.counts):
            table.add_row(value, dist, f"{count:,}")

        console.print(table)
        console.print()


def write_output(summaries: List[SummarySpec], report: Optional[str] = None):
    if report is None:
        return

    mkdir_p(report)
    with smart_open.open(f"{report}/summaries.jsonl.gz", "w") as f:
        for summary in summaries:
            f.write(msgspec.json.encode(summary).decode("utf-8") + "\n")


def create_and_run_analyzer(
    attributes: List[str],
    summaries_path: Optional[str] = None,
    metadata_path: Optional[str] = None,
    report: Optional[str] = None,
    debug: bool = False,
    seed: int = 0,
    num_bins: int = 1000,
    num_processes: int = 1,
    name_regex: Optional[str] = None,
):
    """ """

    # create the report directory if it doesn't exist
    if report:
        mkdir_p(report)

    with ExitStack() as stack:
        # use temporary directories if no paths are provided
        summaries_path = summaries_path or stack.enter_context(TemporaryDirectory())
        metadata_path = metadata_path or stack.enter_context(TemporaryDirectory())

        # make sure these locations exist
        mkdir_p(summaries_path)
        mkdir_p(metadata_path)

        try:
            analyzer = AnalyzerProcessor(
                source_prefix=attributes,
                destination_prefix=summaries_path,
                metadata_prefix=metadata_path,
                debug=debug,
                seed=seed,
                ignore_existing=True,
                retries_on_error=0,
                num_processes=num_processes,
            )
            analyzer(num_bins=num_bins, name_regex=name_regex)

            summaries = aggregate_summaries(summaries_path=summaries_path, num_bins=num_bins)
            visualize_summaries(summaries=summaries)
            write_output(summaries=summaries, report=report)

        finally:
            shutil.rmtree(summaries_path)
            shutil.rmtree(metadata_path)



import argparse
import hashlib
import json
import multiprocessing
import re
from contextlib import ExitStack
from itertools import product
from queue import Queue
from tempfile import TemporaryDirectory
from typing import Any, Dict, List, Tuple, Union

import msgspec
import smart_open
import uniseg.wordbreak

from dolma.core.data_types import InputSpec
from dolma.core.parallel import BaseParallelProcessor
from dolma.core.paths import join_path
from dolma.core.runtime import _make_paths_from_prefix

# we like higher digits first for consistency with previous sampling strategies
HEX_DIGITS = list("0123456789abcdef")[::-1]


def calculate_md5_suffix(prob: float) -> List[str]:
    assert 0 < prob < 1, "Sampling probability must be between 0 and 1 exclusive"

    # calculate the closest fraction of 16 to the probability
    n = 16 * prob

    if n > 1:
        suffix = HEX_DIGITS[: int(round(n))]
    else:
        # not enough hex positions to represent the probability
        # so we use more digits
        more_suffix = calculate_md5_suffix(prob * 16)
        suffix = ["".join(e) for e in product(HEX_DIGITS[:1], more_suffix)]

    assert all(len(s) < 32 for s in suffix), "Suffixes must be less than 32 characters"

    return suffix


class HashSampler(BaseParallelProcessor):
    @classmethod
    def increment_progressbar(  # type: ignore
        cls,
        queue: Queue[Union[Tuple[int, ...], None]],
        /,
        files: int = 0,
        documents: int = 0,
        words: int = 0,
        extracted: int = 0,
    ) -> Dict[str, int]:
        return super().increment_progressbar(
            queue, files=files, documents=documents, extracted=extracted, uniseg_words=words
        )

    @classmethod
    def process_single(
        cls, source_path: str, destination_path: str, queue: "Queue[Union[Tuple[int, ...], None]]", **kwargs: Any
    ):
        decoder = msgspec.json.Decoder(InputSpec)

        probability = kwargs.get("probability", None)
        if probability is None:
            raise ValueError("Probability must be specified")
        probability = float(probability)

        complement = kwargs.get("complement", False)

        suffixes = calculate_md5_suffix(probability)
        re_suffixes = re.compile(rf'({"|".join(suffixes)})$')
        extracted_count = documents_count = words_count = 0
        update_interval = 1

        with ExitStack() as stack:
            source = stack.enter_context(smart_open.open(source_path, "rt"))
            destination = stack.enter_context(smart_open.open(destination_path, "wt"))

            for line in source:
                data = decoder.decode(line)
                md5 = hashlib.md5(data.text.encode()).hexdigest()

                if complement and not re_suffixes.search(md5):
                    extracted_count += 1
                    words_count += sum(1 for w in uniseg.wordbreak.words(data.text.strip()) if w.strip())
                    destination.write(line)
                elif not complement and re_suffixes.search(md5):
                    extracted_count += 1
                    words_count += sum(1 for w in uniseg.wordbreak.words(data.text.strip()) if w.strip())
                    destination.write(line)

                documents_count += 1

                if documents_count % update_interval == 0:
                    # update the progress bar every 1000 documents to prevent
                    # buffering
                    cls.increment_progressbar(
                        queue, documents=documents_count, extracted=extracted_count, words=words_count
                    )
                    extracted_count = documents_count = words_count = 0

                    if queue.qsize() >= multiprocessing.cpu_count():
                        # double the update interval if the queue is full
                        update_interval *= 2

        cls.increment_progressbar(
            queue, files=1, documents=documents_count, extracted=extracted_count, words=words_count
        )


def main(
    source: Union[List[str], str],
    destination: str,
    probability: float,
    complement: bool = False,
    num_workers: int = 1,
    debug: bool = False,
    dryrun: bool = False,
) -> None:
    # make source always a list
    source = [source] if isinstance(source, str) else source

    pattern = calculate_md5_suffix(probability)
    if complement:
        print(f"Sampling complement of probability {probability} split by excluding MD5 suffixes {pattern}")
    else:
        print(f"Sampling with probability {probability} using MD5 suffixes {pattern}")

    if dryrun:
        return

    with TemporaryDirectory() as tempdir:
        if len(source) > 1:
            dest_prefixes = _make_paths_from_prefix(source, join_path(None, destination))
            meta_prefixes = _make_paths_from_prefix(source, join_path(None, tempdir))
        else:
            dest_prefixes = [destination]
            meta_prefixes = [tempdir]

        processor = HashSampler(
            source_prefix=source,
            destination_prefix=dest_prefixes,
            metadata_prefix=meta_prefixes,
            num_processes=num_workers,
            debug=debug,
        )
        processor(probability=probability, complement=complement)


def parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument("-s", "--source", type=str, required=True, help="glob pattern for source files", nargs="+")
    ap.add_argument("-d", "--destination", type=str, required=True, help="destination prefix")
    ap.add_argument("-p", "--probability", type=float, required=True, help="sampling probability")
    ap.add_argument("-n", "--num-workers", type=int, default=1, help="number of workers")
    ap.add_argument("--debug", action="store_true", help="debug mode")
    ap.add_argument("--dryrun", action="store_true", help="dry run")
    ap.add_argument("--complement", action="store_true", help="get the complement of the sample")
    opts = ap.parse_args()

    print(json.dumps(vars(opts), indent=2, sort_keys=True))
    return opts


if __name__ == "__main__":
    opts = parse_args()
    main(**vars(opts))
